#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from .SequentialInferenceDistributions import AR1TransitionDistribution
from TransportMaps.Distributions.FrozenDistributions import \
    GaussianDistribution
from TransportMaps.Distributions.ConditionalDistributions import \
    ConditionallyGaussianDistribution
from TransportMaps.Maps.MapBase import \
    Map, LinearMap, ConstantMap, ConditionallyLinearMap

__all__ = ['LinearGaussianAR1TransitionDistribution',
           'ConditionallyLinearGaussianAR1TransitionDistribution']

class LinearGaussianAR1TransitionDistribution(AR1TransitionDistribution):
    r""" Transition probability distribution :math:`g({\bf x}_{k-1},{\bf x}_k) = \pi_{{\bf X}_k \vert {\bf X}_{k-1}={\bf x}_{k-1}}({\bf x}_k) = \pi({\bf x}_k - F_k {\bf x}_{k-1} - {\bf c}_k)` where :math:`\pi \sim \mathcal{N}(\mu_k,Q_k)`.

    This represents the following Markov transition model:

    .. math::

       {\bf x}_k = c_k + F_k {\bf x}_{k-1} + {\bf w}_k \\
       {\bf w}_k \sim \mathcal{N}(\mu,Q_k)

    where the control :math:`{\bf c}_k := B_k {\bf u}_k` can be used for control purposes
    
    Args:
      ck (:class:`ndarray<numpy.ndarray>` [:math:`d`] or :class:`Map<TransportMaps.Maps.Map>`): constant part or map returning the constant part given some parameters
      Fk (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): state transition matrix (dynamics) or map returning the linear part given some parametrs
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d`] or :class:`Map<TransportMaps.Maps.Map>`): mean :math:`\mu_k` or parametric map for :math:`\mu_k(\theta)`
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): covariance :math:`Q_k` or parametric map for :math:`Q_k(\theta)`
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): precision :math:`Q_k^{-1}` or parametric map for :math:`Q_k^{-1}(\theta)`
    """
    def __init__(self, ck, Fk, mu, sigma=None, precision=None,
                 coeffs=None):
        Fmap = LinearMap(ck, Fk)
        pi = GaussianDistribution(mu, sigma=sigma, precision=precision)
        super(LinearGaussianAR1TransitionDistribution, self).__init__(
            pi, Fmap)

class ConditionallyLinearGaussianAR1TransitionDistribution(AR1TransitionDistribution):
    r""" Transition probability distribution :math:`g(\theta,{\bf x}_{k-1},{\bf x}_k) = \pi_{{\bf X}_k \vert {\bf X}_{k-1}={\bf x}_{k-1}}({\bf x}_k, \Theta=\theta) = \pi({\bf x}_k - F_k(\theta) {\bf x}_{k-1} - {\bf c}_k(\theta))` where :math:`\pi \sim \mathcal{N}(\mu_k(\theta),Q_k(\theta))`.

    This represents the following Markov transition model:

    .. math::

       {\bf x}_k = c_k + F_k {\bf x}_{k-1} + {\bf w}_k \\
       {\bf w}_k \sim \mathcal{N}(\mu,Q_k)

    where the control :math:`{\bf c}_k := B_k {\bf u}_k` can be used for control purposes
    
    Args:
      ck (:class:`ndarray<numpy.ndarray>` [:math:`d`] or :class:`Map<TransportMaps.Maps.Map>`): constant part or map returning the constant part given some parameters
      Fk (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): state transition matrix (dynamics) or map returning the linear part given some parametrs
      mu (:class:`ndarray<numpy.ndarray>` [:math:`d`] or :class:`Map<TransportMaps.Maps.Map>`): mean :math:`\mu_k` or parametric map for :math:`\mu_k(\theta)`
      sigma (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): covariance :math:`Q_k` or parametric map for :math:`Q_k(\theta)`
      precision (:class:`ndarray<numpy.ndarray>` [:math:`d,d`] or :class:`Map<TransportMaps.Maps.Map>`): precision :math:`Q_k^{-1}` or parametric map for :math:`Q_k^{-1}(\theta)`
      coeffs (:class:`ndarray<numpy.ndarray>`): fixing the coefficients :math:`\theta`
    """
    def __init__(self, ck, Fk, mu, sigma=None, precision=None, coeffs=None):
        # DISTRIBUTION
        if isinstance(mu, np.ndarray) and (
                (sigma is not None and isinstance(sigma, np.ndarray)) or
                (precision is not None and isinstance(precision, np.ndarray)) ):
            pi = GaussianDistribution(mu, sigma=sigma, precision=precision)
        else:
            pi = ConditionallyGaussianDistribution(mu, sigma=sigma, precision=precision)
        Fmap = ConditionallyLinearMap(ck, Fk)
        super(ConditionallyLinearGaussianAR1TransitionDistribution, self).__init__(
            pi, Fmap)
        self.coeffs = coeffs

    @property
    def n_coeffs(self):
        return self._n_coeffs
        
    @property
    def coeffs(self):
        return self._coeffs

    @coeffs.setter
    def coeffs(self, coeffs):
        if coeffs is not None:
            self.T.coeffs = coeffs
            if self.isPiCond:
                self.pi.coeffs = coeffs
            self._coeffs = coeffs


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import scipy.stats as stats

from TransportMaps.Functionals.MonotonicFunctionApproximations import *

__all__ = ['MonotonicFrozenFunction', 'FrozenLinear',
           'FrozenExponential', 'FrozenGaussianToUniform']

class MonotonicFrozenFunction(MonotonicFunctionApproximation):
    r""" [Abstract] Frozen function. No optimization over the coefficients allowed.
    """
    def precomp_evaluate(self, x, *args, **kwargs):
        return {}

    def precomp_grad_x(self, x, *args, **kwargs):
        return {}

    def precomp_hess_x(self, x, *args, **kwargs):
        return {}

    def precomp_partial_xd(self, x, *args, **kwargs):
        return {}

    def precomp_grad_x_partial_xd(self, x, *args, **kwargs):
        return {}

    def precomp_hess_x_partial_xd(self, x, *args, **kwargs):
        return {}

    def precomp_partial2_xd(self, x, *args, **kwargs):
        return {}

class FrozenLinear(MonotonicFrozenFunction):
    r""" Frozen Linear map :math:`{\bf x} \rightarrow a_1 + a_2 {\bf x}_d`

    Args:
      dim (int): input dimension :math:`d`
      a1 (int): coefficient :math:`a_1`
      a2 (int): coefficient :math:`a_2`
    """

    def __init__(self, dim, a1, a2):
        self.dim = dim
        self.set_coeffs(a1, a2)

    @property
    def n_coeffs(self):
        r"""
        Returns: 2
        """
        return 2

    def set_coeffs(self, a1, a2):
        r""" Set coefficients :math:`a_1` and :math:`a_2`. """
        if a2 < 0.:
            raise ValueError("The map is not monotone")
        self.a1 = a1
        self.a2 = a2

    def evaluate(self, x, *args, **kwargs):
        r""" Evaluate :math:`f_{\bf a}({\bf x}) = a_1 + a_2 {\bf x}_d`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function value at points `x`
        """
        return self.a1 + self.a2 * x[:,-1]

    def grad_x(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla_{\bf x} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial_{{\bf x}_1} f_{\bf a}({\bf x}) \\ \partial_{{\bf x}_2} f_{\bf a}({\bf x}) \\
           \vdots \\ \partial_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix} 0 \\ 0 \\ \vdots \\ a2 \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}f_{\bf a}({\bf x};{\bf a})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim))
        out[:,-1] = self.a2
        return out

    def partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial_{{\bf x}_d} f_{\bf a}({\bf x}) = b`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return self.a2 * np.ones(x.shape[0])

    def grad_x_partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\partial_{{\bf x}_d} f_{\bf a}({\bf x}) = 0`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return np.zeros((x.shape[0],self.dim))

    def partial2_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial^2_{{\bf x}_d} f_{\bf a}({\bf x}) = 0`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{{\bf x}_d}f_{\bf a}({\bf x};{\bf a})` at points `x`
        """
        return np.zeros(x.shape[0])

    def inverse(self, xkm1, y):
        r""" Compute :math:`f_{\bf a}^{-1}({\bf x}_{1:d-1})(y):={\bf x}_d` s.t. :math:`f_{\bf a}({\bf x}_{1:d-1},{\bf x}_d) - y = 0`.

        Due to the form of the approximation we have:

        .. math::

           f_{\bf a}^{-1}({\bf x}_{1:d-1})(y) = \frac{y-a_1}{a_2}

        Args:
          xkm1 (:class:`ndarray<numpy.ndarray>` [:math:`d-1`]): fixed coordinates
            :math:`{\bf x}_{1:d-1}`
          y (float): value :math:`y`

        Returns:
          (:class:`float<float>`) -- inverse value :math:`x`.
        """
        return (y - self.a1)/self.a2

class FrozenExponential(MonotonicFrozenFunction):
    r""" Frozen Exponential map :math:`f_{\bf a}:{\bf x} \mapsto \exp( {\bf x}_d )`

    Args:
      dim (int): input dimension :math:`d`
    """

    def __init__(self, dim):
        self.dim = dim
        self.set_coeffs()

    @property
    def n_coeffs(self):
        r"""
        Returns: 0
        """
        return 0

    def set_coeffs(self):
        r""" No coefficients to be set.
        """
        pass

    def evaluate(self, x, *args, **kwargs):
        r""" Evaluate :math:`f_{\bf a}({\bf x}) = \exp({\bf x}_d)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function value at points `x`
        """
        return np.exp(x[:,-1])

    def grad_x(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla_{\bf x} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial_{{\bf x}_1} f_{\bf a}({\bf x}) \\ \partial_{{\bf x}_2} f_{\bf a}({\bf x}) \\
           \vdots \\ \partial_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix} 0 \\ 0 \\ \vdots \\ \exp({\bf x}_d) \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}f_{\bf a}({\bf x};{\bf a})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim))
        out[:,-1] = np.exp(x[:,-1])
        return out

    def hess_x(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla^2_{\bf x} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial^2_{{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_1{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_1{\bf x}_d} f_{\bf a}({\bf x}) \\
           \partial_{{\bf x}_2 {\bf x}_1} f_{\bf a}({\bf x}) & \partial^2_{{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_2{\bf x}_d} f_{\bf a}({\bf x}) \\
           \vdots &  & \ddots & \\
           \partial_{{\bf x}_d{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_d{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial^2_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix}
           0 & \cdots & 0 & 0 \\
           \vdots & \ddots & 0 & \vdots \\
           0 & \cdots & 0 & 0 \\
           0 & \cdots & 0 &  \exp({\bf x}_d) \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}f_{\bf a}({\bf x})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim,self.dim))
        out[:,-1,-1] = np.exp(x[:,-1])
        return out

    def partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial_{{\bf x}_d} f_{\bf a}({\bf x}) = \exp({\bf x}_d)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{{\bf x}_d}f_{\bf a}({\bf x};{\bf a})` at points `x`
        """
        return self.fapprox(x)

    def grad_x_partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\partial_{{\bf x}_d} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla_{\bf x} \partial_{{\bf x}_d} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial_{{\bf x}_1} f_{\bf a}({\bf x}) \\ \partial_{{\bf x}_2} f_{\bf a}({\bf x}) \\
           \vdots \\ \partial_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix} 0 \\ 0 \\ \vdots \\ \exp({\bf x}_d) \end{bmatrix}


        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return self.dx_fapprox(x)

    def partial2_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial^2_{{\bf x}_d} f_{\bf a}({\bf x}) = \exp({\bf x}_d)`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return self.fapprox(x)

    def inverse(self, xkm1, y):
        r""" Compute :math:`f_{\bf a}^{-1}({\bf x}_{1:d-1})(y):={\bf x}_d` s.t. :math:`f_{\bf a}({\bf x}_{1:d-1},{\bf x}_d) - y = 0`.

        Due to the form of the approximation we have:

        .. math::

           f_{\bf a}^{-1}({\bf x}_{1:d-1})(y) = \log(y)

        Args:
          xkm1 (:class:`ndarray<numpy.ndarray>` [:math:`d-1`]): fixed coordinates
            :math:`{\bf x}_{1:d-1}`
          y (float): value :math:`y`

        Returns:
          (:class:`float<float>`) -- inverse value :math:`x`.
        """
        return np.log(y)

class FrozenGaussianToUniform(MonotonicFrozenFunction):
    r""" Frozen Gaussian To Uniform map.

    This is given by the Cumulative Distribution Function of a standard
    normal distribution along the last coordinate:

    .. math::

       f_{\bf a}({\bf x}) = \frac{1}{2} \left[ 1 + \text{erf}\left( \frac{x}{\sqrt{2}} \right)\right]

    Args:
      dim (int): input dimension :math:`d`
    """

    def __init__(self, dim):
        self.dim = dim
        self.set_coeffs()
        self.std = stats.norm()

    @property
    def n_coeffs(self):
        r"""
        Returns: 0
        """
        return 0

    def set_coeffs(self):
        r""" No coefficients to be set.
        """
        pass

    def evaluate(self, x, *args, **kwargs):
        r""" Evaluate :math:`f_{\bf a}({\bf x})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) -- function value at points `x`
        """
        return self.std.cdf(x[:,-1])

    def grad_x(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla_{\bf x} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial_{{\bf x}_1} f_{\bf a}({\bf x}) \\ \partial_{{\bf x}_2} f_{\bf a}({\bf x}) \\
           \vdots \\ \partial_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix} 0 \\ 0 \\ \vdots \\
           (2\pi)^{-1}\exp(-\frac{{\bf x}^2_d}{2})
           \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}f_{\bf a}({\bf x})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim))
        out[:,-1] = np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)
        return out

    def partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial_{{\bf x}_d} f_{\bf a}({\bf x}) = (2\pi)^{-1}\exp(-\frac{{\bf x}^2_d}{2})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)

    def hess_x(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla^2_{\bf x} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial^2_{{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_1{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_1{\bf x}_d} f_{\bf a}({\bf x}) \\
           \partial_{{\bf x}_2 {\bf x}_1} f_{\bf a}({\bf x}) & \partial^2_{{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_2{\bf x}_d} f_{\bf a}({\bf x}) \\
           \vdots &  & \ddots & \\
           \partial_{{\bf x}_d{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_d{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial^2_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix}
           0 & \cdots & 0 & 0 \\
           \vdots & \ddots & 0 & \vdots \\
           0 & \cdots & 0 & 0 \\
           0 & \cdots & 0 &  -{\bf x}_d (2\pi)^{-1} \exp(-\frac{{\bf x}^2_d}{2})
           \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}f_{\bf a}({\bf x})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim,self.dim))
        out[:,-1,-1] = -x[:,-1] * np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)
        return out

    def grad_x_partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla_{\bf x}\partial_{{\bf x}_d} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla_{\bf x} \partial_{{\bf x}_d} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial_{{\bf x}_1} f_{\bf a}({\bf x}) \\ \partial_{{\bf x}_2} f_{\bf a}({\bf x}) \\
           \vdots \\ \partial_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix} 0 \\ 0 \\ \vdots \\
           -{\bf x}_d (2\pi)^{-1} \exp(-\frac{{\bf x}^2_d}{2})
           \end{bmatrix}


        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]) --
            :math:`\nabla_{\bf x}\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim))
        out[:,-1] = - x[:,-1] * np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)
        return out

    def partial2_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\partial^2_{{\bf x}_d} f_{\bf a}({\bf x}) = -{\bf x}_d (2\pi)^{-1} \exp(-\frac{{\bf x}^2_d}{2})`

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m`]) --
            :math:`\partial^2_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        return - x[:,-1] * np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)

    def hess_x_partial_xd(self, x, *args, **kwargs):
        r""" Evaluate :math:`\nabla^2_{\bf x} \partial_{{\bf x}_d} f_{\bf a}({\bf x})`

        This is:

        .. math::

           \nabla^2_{\bf x} \partial_{{\bf x}_d} f_{\bf a}({\bf x}) = \begin{bmatrix}
           \partial^2_{{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_1{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_1{\bf x}_d} f_{\bf a}({\bf x}) \\
           \partial_{{\bf x}_2 {\bf x}_1} f_{\bf a}({\bf x}) & \partial^2_{{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial_{{\bf x}_2{\bf x}_d} f_{\bf a}({\bf x}) \\
           \vdots &  & \ddots & \\
           \partial_{{\bf x}_d{\bf x}_1} f_{\bf a}({\bf x}) & \partial_{{\bf x}_d{\bf x}_2} f_{\bf a}({\bf x}) & \cdots & \partial^2_{{\bf x}_d} f_{\bf a}({\bf x})
           \end{bmatrix} = \begin{bmatrix}
           0 & \cdots & 0 & 0 \\
           \vdots & \ddots & 0 & \vdots \\
           0 & \cdots & 0 & 0 \\
           0 & \cdots & 0 & ({\bf x}_d - 1) (2\pi)^{-1} \exp(-\frac{{\bf x}^2_d}{2})
           \end{bmatrix}

        Args:
          x (:class:`ndarray<numpy.ndarray>` [:math:`m,d`]): evaluation points.

        Returns:
          (:class:`ndarray<numpy.ndarray>` [:math:`m,d,d`]) --
            :math:`\nabla^2_{\bf x}\partial_{{\bf x}_d}f_{\bf a}({\bf x})` at points `x`
        """
        out = np.zeros((x.shape[0],self.dim,self.dim))
        out[:,-1,-1] = (x[:,-1]**2. - 1.) * np.exp(-x[:,-1]**2./2.) / np.sqrt(2.*np.pi)
        return out

    def invserse(self, xkm1, y):
        r""" Compute :math:`f_{\bf a}^{-1}({\bf x}_{1:d-1})(y):={\bf x}_d` s.t. :math:`f_{\bf a}({\bf x}_{1:d-1},{\bf x}_d) - y = 0`.

        Due to the form of the approximation we have:

        .. math::

           f_{\bf a}^{-1}({\bf x}_{1:d-1})(y) = \sqrt{2} \text{erf}^{-1}(2y-1)

        Args:
          xkm1 (:class:`ndarray<numpy.ndarray>` [:math:`d-1`]): fixed coordinates
            :math:`{\bf x}_{1:d-1}`
          y (float): value :math:`y`

        Returns:
          (:class:`float<float>`) -- inverse value :math:`x`.
        """
        return self.std.ppf( y )



#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np
import scipy.linalg as scila
import scipy.integrate as sciint

import SpectralToolbox.Spectral1D as S1D

from TransportMaps.Maps.MapBase import Map, ConstantMap
from TransportMaps.Distributions import \
    GaussianDistribution, ConditionallyGaussianDistribution
from TransportMaps.Distributions.Decomposable.LinearGaussianSequentialInferenceDistributions import \
    ConditionallyLinearGaussianAR1TransitionDistribution
from TransportMaps.Likelihoods import \
    AdditiveConditionallyLinearGaussianLogLikelihood

__all__ = ['DEFAULTS',
           'Coradia175Vehicle',
           'DynamicsMap', 'NoiseDynamicsMap', 'ObservationSystemMap',
           'ParametersPrior', 'StateSpacePrior', 'StateSpaceTransition',
           'StateSpaceLogLikelihood', 'generate_data',
           'Yw1', 'Yw1dot',
           'Pw1', 'Pw1dot', 
           'Yw2', 'Yw2dot', 
           'Pw2', 'Pw2dot', 
           'Yb', 'Ybdot',
           'Pb', 'Pbdot',  
           'Yc', 'Ycdot',  
           'D1', 'D2']

nax = np.newaxis

# Position of degrees of freedom
NDOFS  = 0
Yw1    = NDOFS; NDOFS += 1;
Yw1dot = NDOFS; NDOFS += 1;
Pw1    = NDOFS; NDOFS += 1;
Pw1dot = NDOFS; NDOFS += 1;
Yw2    = NDOFS; NDOFS += 1;
Yw2dot = NDOFS; NDOFS += 1;
Pw2    = NDOFS; NDOFS += 1;
Pw2dot = NDOFS; NDOFS += 1;
Yb     = NDOFS; NDOFS += 1;
Ybdot  = NDOFS; NDOFS += 1;
Pb     = NDOFS; NDOFS += 1;
Pbdot  = NDOFS; NDOFS += 1;
Yc     = NDOFS; NDOFS += 1;
Ycdot  = NDOFS; NDOFS += 1;
D1     = NDOFS; NDOFS += 1;
D2     = NDOFS; NDOFS += 1;

# Default values
DEFAULTS = {
    'L_Ky':   (4000.)*10**3,   # Leading primary spring lateral (kN/m)
    'T_Ky':   (4000.)*10**3,   # Trailing primary spring lateral (kN/m)
    'L_KPsi': (4000.)*10**3,   # Primary yaw stiffness (kN/rad)
    'T_KPsi': (4000.)*10**3,   # Primary yaw stiffness (kN/rad)
    'Kyb':    (160.)*10**3,    # Secondary lateral stiffness (kN/m)
    'Cyb':    (16.)*10**3,     # Secondary lateral damping (kN s/m)
    'CPb':    (500.)*10**3,    # Secondary anti-yaw damping (kN s/rad)
    'f11':    7.4*10**6,       # Longitudinal creep coeff (MN)
    'f22':    6.2*10**6,       # Lateral creep coeff (MN)
}


class Coradia175Vehicle(object):
    r"""
    Args:
      v (float): Longitudinal vehicle velocity (m/s)
    """
    mw     = 1000.     # Wheel set mass (kg)
    Iw     = 600.      # Wheel set yaw inertia (kg * m^2)
    mb     = 2000.     # Bogie mass (kg)
    Ib     = 2300.     # Bogie yaw inertia (kg * m^2)
    mc     = 8720.     # Half vehicle mass (kg)
    a      = 1.05      # Semi wheel-wheel spacing (m)
    l      = 0.75      # Half gauge (m)
    lmb    = 0.15      # Conicity
    r0     = 0.37      # Wheelset radius (m)
    
    def __init__(self, v,
                 observables=[Ybdot, Pb, Ycdot],
                 dt = 1e-2,                        # Discretization time
                 # Ar = 2 * 0.33 * 1e-3 # Original
                 Ar = 2 * 0.33 * 1e-6,             # Track roughness - reduced
                 # Noises
                 init_noise = 1e-6,
                 # obs_noise = {'Yw1dot': 1e-4 * 1e2,
                 #              'Yw2dot': 1e-4 * 1e2,
                 #              'Ybdot':  4.11 * 1e-4 * 1e2,
                 #              'Pb':     3.75 * 1e-7 * 1e2,
                 #              'Ycdot':  7.62 * 1e-3 * 1e2}
                 obs_noise = {'Yw1dot': 1e-4 * 1e1,
                              'Yw2dot': 1e-4 * 1e1,
                              'Ybdot':  4.11 * 1e-4 * 1e1,
                              'Pb':     3.75 * 1e-7 * 1e1,
                              'Ycdot':  7.62 * 1e-3 * 1e1}
                 # obs_noise = {'Yw1dot': 1e-4,
                 #              'Yw2dot': 1e-4,
                 #              'Ybdot':  4.11 * 1e-4,
                 #              'Pb':     3.75 * 1e-7,
                 #              'Ycdot':  7.62 * 1e-3}
             ):
        self.v      = v         # Vehicle forward velocity (m/s)
        self.observables = observables
        self.dt = dt
        self.Ar = Ar
        self.init_noise = init_noise
        self.obs_noise = obs_noise

    def get_system_matrix(
            self,
            L_Ky   = DEFAULTS['L_Ky'],     # Leading primary spring lateral (kN/m)
            T_Ky   = DEFAULTS['T_Ky'],     # Trailing primary spring lateral (kN/m)
            L_KPsi = DEFAULTS['L_KPsi'],   # Primary yaw stiffness (kN/rad)
            T_KPsi = DEFAULTS['T_KPsi'],   # Primary yaw stiffness (kN/rad)
            Kyb    = DEFAULTS['Kyb'],      # Secondary lateral stiffness (kN/m)
            Cyb    = DEFAULTS['Cyb'],      # Secondary lateral damping (kN s/m)
            CPb    = DEFAULTS['CPb'],      # Secondary anti-yaw damping (kN s/rad)
            f11    = DEFAULTS['f11'],      # Longitudinal creep coeff (MN)
            f22    = DEFAULTS['f22'],      # Lateral creep coeff (MN)
    ):
        ###################
        # System dynamics #
        ###################
        A = np.zeros((NDOFS,NDOFS))
        # Lateral displacement front wheelset
        A[Yw1,Yw1dot]    = 1.
        A[Yw1dot,Yw1]    = - L_Ky
        A[Yw1dot,Yw1dot] = - 2. * f22 / self.v
        A[Yw1dot,Pw1]    = 2. * f22 
        A[Yw1dot,Yb]     = L_Ky 
        A[Yw1dot,Pb]     = self.a * L_Ky 
        A[Yw1dot,:]     /= self.mw
        # Yaw front wheelset
        A[Pw1,Pw1dot]    = 1.
        A[Pw1dot,Pw1dot] = - 2. * f11 * self.l**2 / self.v
        A[Pw1dot,Pw1]    = - L_KPsi
        A[Pw1dot,Pb]     = L_KPsi
        A[Pw1dot,D1]     = - 2. * f11 * self.lmb * self.l / self.r0
        A[Pw1dot,:]     /= self.Iw
        # Lateral displacement trailing wheelset
        A[Yw2,Yw2dot]    = 1.
        A[Yw2dot,Yw2]    = - T_Ky
        A[Yw2dot,Yw2dot] = - 2. * f22 / self.v
        A[Yw2dot,Pw2]    = 2. * f22 
        A[Yw2dot,Yb]     = T_Ky 
        A[Yw2dot,Pb]     = self.a * T_Ky 
        A[Yw2dot,:]     /= self.mw
        # Yaw trailing wheelset
        A[Pw2,Pw2dot]    = 1.
        A[Pw2dot,Pw2dot] = - 2. * f11 * self.l**2 / self.v
        A[Pw2dot,Pw2]    = - T_KPsi
        A[Pw2dot,Pb]     = T_KPsi
        A[Pw2dot,D2]     = - 2. * f11 * self.lmb * self.l / self.r0
        A[Pw2dot,:]     /= self.Iw
        # Lateral displacement bogie frame
        A[Yb,Ybdot]      = 1.
        A[Ybdot,Yw1]     = L_Ky
        A[Ybdot,Yw2]     = T_Ky
        A[Ybdot,Ybdot]   = - Cyb
        A[Ybdot,Yb]      = - (L_Ky + T_Ky + Kyb)
        A[Ybdot,Ycdot]   = Cyb
        A[Ybdot,Yc]      = Kyb
        A[Ybdot,:]      /= self.mb
        # Yaw bogie frame
        A[Pb,Pbdot]      = 1.
        A[Pbdot,Yw1]     = self.a * L_Ky
        A[Pbdot,Pw1]     = L_KPsi
        A[Pbdot,Yw2]     = - self.a * T_Ky 
        A[Pbdot,Pw2]     = T_KPsi
        A[Pbdot,Pbdot]   = - CPb
        A[Pbdot,Pb]      = - (L_Ky + T_Ky) * self.a**2 \
                                - (L_KPsi + T_KPsi)
        A[Pbdot,:]      /= self.Ib
        # Lateral displacement car body (MODIFIED!)
        A[Yc,Ycdot]      = 1.
        A[Ycdot,Ybdot]   = Cyb
        A[Ycdot,Yb]      = Kyb
        A[Ycdot,Ycdot]   = -Cyb
        A[Ycdot,Yc]      = - Kyb
        A[Ycdot,:]      /= self.mc
        # Lateral track displacement (leading wheelset)
        A[D1,Yw1dot]     = 1.
        A[D2,Yw2dot]     = 1.
        return A

    def get_grad_system_matrix(
            self,
            par_name_list,  # List of parameters over which to compute the gradient
            L_Ky   = DEFAULTS['L_Ky'],     # Leading primary spring lateral (kN/m)
            T_Ky   = DEFAULTS['T_Ky'],     # Trailing primary spring lateral (kN/m)
            L_KPsi = DEFAULTS['L_KPsi'],   # Primary yaw stiffness (kN/rad)
            T_KPsi = DEFAULTS['T_KPsi'],   # Primary yaw stiffness (kN/rad)
            Kyb    = DEFAULTS['Kyb'],      # Secondary lateral stiffness (kN/m)
            Cyb    = DEFAULTS['Cyb'],      # Secondary lateral damping (kN s/m)
            CPb    = DEFAULTS['CPb'],      # Secondary anti-yaw damping (kN s/rad)
            f11    = DEFAULTS['f11'],      # Longitudinal creep coeff (MN)
            f22    = DEFAULTS['f22'],      # Lateral creep coeff (MN)
    ):
        gA = np.zeros((NDOFS,NDOFS,len(par_name_list)))
        cntr = 0 # Counter for free parameters
        if 'L_Ky' in par_name_list:
            # Lateral displacement front wheelset
            gA[Yw1dot,Yw1,cntr]    = - 1.
            gA[Yw1dot,Yb,cntr]     = 1.
            gA[Yw1dot,Pb,cntr]     = self.a
            gA[Yw1dot,:,cntr]     /= self.mw
            # Lateral displacement bogie frame
            gA[Ybdot,Yw1,cntr]     = 1.
            gA[Ybdot,Yb,cntr]      = - 1.
            gA[Ybdot,:,cntr]      /= self.mb
            # Yaw bogie frame
            gA[Pbdot,Yw1,cntr]     = self.a 
            gA[Pbdot,Pb,cntr]      = - self.a**2 
            gA[Pbdot,:,cntr]      /= self.Ib
            # Update counter
            cntr += 1
        if 'T_Ky' in par_name_list:
            # Lateral displacement trailing wheelset
            gA[Yw2dot,Yw2,cntr]    = - 1.
            gA[Yw2dot,Yb,cntr]     = 1.
            gA[Yw2dot,Pb,cntr]     = self.a 
            gA[Yw2dot,:,cntr]     /= self.mw
            # Lateral displacement bogie frame
            gA[Ybdot,Yw2,cntr]     = 1.
            gA[Ybdot,Yb,cntr]      = - 1.
            gA[Ybdot,:,cntr]      /= self.mb
            # Yaw bogie frame
            gA[Pbdot,Yw2,cntr]     = - self.a 
            gA[Pbdot,Pb,cntr]      = - self.a**2 
            gA[Pbdot,:,cntr]      /= self.Ib
            # Update counter
            cntr += 1
        if 'L_KPsi' in par_name_list:
            # Yaw front wheelset
            gA[Pw1dot,Pw1,cntr]    = - 1.
            gA[Pw1dot,Pb,cntr]     = 1.
            gA[Pw1dot,:,cntr]     /= self.Iw
            # Yaw bogie frame
            gA[Pbdot,Pw1,cntr]     = 1.
            gA[Pbdot,Pb,cntr]      = - 1.
            gA[Pbdot,:,cntr]      /= self.Ib
            # Update counter
            cntr += 1
        if 'T_KPsi' in par_name_list:
            # Yaw trailing wheelset
            gA[Pw2dot,Pw2,cntr]    = - 1.
            gA[Pw2dot,Pb,cntr]     = 1.
            gA[Pw2dot,:,cntr]     /= self.Iw
            # Yaw bogie frame
            gA[Pbdot,Pw2,cntr]     = 1.
            gA[Pbdot,Pb,cntr]      = - 1.
            gA[Pbdot,:,cntr]      /= self.Ib
            # Update counter
            cntr += 1
        if 'Kyb' in par_name_list:
            # Lateral displacement bogie frame
            gA[Ybdot,Yb,cntr]      = - 1.
            gA[Ybdot,Yc,cntr]      = 1.
            gA[Ybdot,:,cntr]      /= self.mb
            # Lateral displacement car body (MODIFIED!)
            gA[Ycdot,Yb,cntr]      = 1.
            gA[Ycdot,Yc,cntr]      = - 1.
            gA[Ycdot,:,cntr]      /= self.mc
            # Update counter
            cntr += 1
        if 'Cyb' in par_name_list:
            # Lateral displacement bogie frame
            gA[Ybdot,Ybdot,cntr]   = - 1.
            gA[Ybdot,Ycdot,cntr]   = 1.
            gA[Ybdot,:,cntr]      /= self.mb
            # Lateral displacement car body (MODIFIED!)
            gA[Ycdot,Ybdot,cntr]   = 1.
            gA[Ycdot,Ycdot,cntr]   = - 1.
            gA[Ycdot,:,cntr]      /= self.mc
            # Update counter
            cntr += 1
        if 'CPb' in par_name_list:
            # Yaw bogie frame
            gA[Pbdot,Pbdot,cntr]   = - 1.
            gA[Pbdot,:,cntr]      /= self.Ib
            # Update counter
            cntr += 1
        if 'f11' in par_name_list:
            # Yaw front wheelset
            gA[Pw1dot,Pw1dot,cntr] = - 2. * self.l**2 / self.v
            gA[Pw1dot,D1,cntr]     = - 2. * self.lmb * self.l / self.r0
            gA[Pw1dot,:,cntr]     /= self.Iw
            # Yaw trailing wheelset
            gA[Pw2dot,Pw2dot,cntr] = - 2. * self.l**2 / self.v
            gA[Pw2dot,D2,cntr]     = - 2. * self.lmb * self.l / self.r0
            gA[Pw2dot,:,cntr]     /= self.Iw
            # Update counter
            cntr += 1
        if 'f22' in par_name_list:
            # Lateral displacement front wheelset
            gA[Yw1dot,Yw1dot,cntr] = - 2. / self.v
            gA[Yw1dot,Pw1,cntr]    = 2.
            gA[Yw1dot,:,cntr]     /= self.mw
            # Lateral displacement trailing wheelset
            gA[Yw2dot,Yw2dot,cntr] = - 2. / self.v
            gA[Yw2dot,Pw2,cntr]    = 2. 
            gA[Yw2dot,:,cntr]     /= self.mw
            # Update counter
            cntr += 1
        return gA
        
    def get_track_matrix(self):
        #################
        # Track control #
        #################
        G = np.zeros((NDOFS,NDOFS))
        G[D1,D1] = -1.
        G[D2,D2] = -1.
        return G

    def get_grad_track_matrix(self, par_name_list):
        dG = np.zeros((NDOFS,NDOFS,len(par_name_list)))
        return dG

    def get_observation_matrix(
            self,
            L_Ky   = DEFAULTS['L_Ky'],     # Leading primary spring lateral (kN/m)
            T_Ky   = DEFAULTS['T_Ky'],     # Trailing primary spring lateral (kN/m)
            L_KPsi = DEFAULTS['L_KPsi'],   # Primary yaw stiffness (kN/rad)
            T_KPsi = DEFAULTS['T_KPsi'],   # Primary yaw stiffness (kN/rad)
            Kyb    = DEFAULTS['Kyb'],      # Secondary lateral stiffness (kN/m)
            Cyb    = DEFAULTS['Cyb'],      # Secondary lateral damping (kN s/m)
            CPb    = DEFAULTS['CPb'],      # Secondary anti-yaw damping (kN s/rad)
            f11    = DEFAULTS['f11'],      # Longitudinal creep coeff (MN)
            f22    = DEFAULTS['f22'],      # Lateral creep coeff (MN)
    ):
        A = self.get_system_matrix(L_Ky, T_Ky, L_KPsi, T_KPsi,
                                   Kyb, Cyb, CPb, f11, f22)
        ################
        # Observations #
        ################
        H = A[self.observables, :]
        return H

    def get_grad_observation_matrix(
            self,
            par_name_list,
            L_Ky   = DEFAULTS['L_Ky'],     # Leading primary spring lateral (kN/m)
            T_Ky   = DEFAULTS['T_Ky'],     # Trailing primary spring lateral (kN/m)
            L_KPsi = DEFAULTS['L_KPsi'],   # Primary yaw stiffness (kN/rad)
            T_KPsi = DEFAULTS['T_KPsi'],   # Primary yaw stiffness (kN/rad)
            Kyb    = DEFAULTS['Kyb'],      # Secondary lateral stiffness (kN/m)
            Cyb    = DEFAULTS['Cyb'],      # Secondary lateral damping (kN s/m)
            CPb    = DEFAULTS['CPb'],      # Secondary anti-yaw damping (kN s/rad)
            f11    = DEFAULTS['f11'],      # Longitudinal creep coeff (MN)
            f22    = DEFAULTS['f22'],      # Lateral creep coeff (MN)
    ):
        gA = self.get_grad_system_matrix(
            par_name_list,
            L_Ky, T_Ky, L_KPsi, T_KPsi, Kyb, Cyb, CPb, f11, f22)
        gH = gA[self.observables, :, :]
        return gH
        
    def get_matrices(
            self,
            L_Ky   = DEFAULTS['L_Ky'],     # Leading primary spring lateral (kN/m)
            T_Ky   = DEFAULTS['T_Ky'],     # Trailing primary spring lateral (kN/m)
            L_KPsi = DEFAULTS['L_KPsi'],   # Primary yaw stiffness (kN/rad)
            T_KPsi = DEFAULTS['T_KPsi'],   # Primary yaw stiffness (kN/rad)
            Kyb    = DEFAULTS['Kyb'],      # Secondary lateral stiffness (kN/m)
            Cyb    = DEFAULTS['Cyb'],      # Secondary lateral damping (kN s/m)
            CPb    = DEFAULTS['CPb'],      # Secondary anti-yaw damping (kN s/rad)
            f11    = DEFAULTS['f11'],      # Longitudinal creep coeff (MN)
            f22    = DEFAULTS['f22'],      # Lateral creep coeff (MN)
    ):
        A = self.get_system_matrix(L_Ky, T_Ky, L_KPsi, T_KPsi,
                                   Kyb, Cyb, CPb, f11, f22)
        G = self.get_track_matrix()
        H = self.get_observation_matrix(L_Ky, T_Ky, L_KPsi, T_KPsi,
                                        Kyb, Cyb, CPb, f11, f22)
        return (A, G, H)

def grad_exp_integrand(xx, A, gA):
    out = np.zeros((xx.shape[0],) + gA.shape)
    for i, x in enumerate(xx):
        out[i] = np.einsum('ij,jk...->ik...', scila.expm(x*A),
                           np.einsum('ij...,jk->ik...', gA, scila.expm((1.-x)*A)) )
    return out

def chebnorm(A):
    return np.max(np.abs(A))

def grad_exp(A, gA, rtol=1e-10, atol=1e-16, maxord=100, startord=-1, noit=False):
    P = S1D.JacobiPolynomial(0.,0.)
    order = startord
    old = np.zeros(gA.shape)
    new = np.inf * np.ones(gA.shape)
    nrm = [chebnorm(old - new)]
    stop = False
    while not stop and order < maxord and \
          nrm[-1] > rtol * min(chebnorm(old),chebnorm(new)) + atol:
        order += 1
        old = new
        (x, w) = P.Quadrature(order)
        x = (x+1)/2.
        w /= 2.
        new = np.einsum('i,i...->...', w, grad_exp_integrand(x, A, gA))
        nrm.append( chebnorm(old - new) )
        stop = noit
    if order == maxord:
        raise RuntimeError("Quadrature did not converge")
    return new

class ParametersPrior(GaussianDistribution):
    def __init__(self, par_name_list):
        self.par_name_list = par_name_list
        dim = len(par_name_list)
        mu = np.zeros(dim)
        sigma2 = np.zeros((dim,dim))
        for i, par_name in enumerate(par_name_list):
            mu[i] = DEFAULTS[par_name]
            std = 0.05 * mu[i]
            sigma2[i,i] = std**2
        super(ParametersPrior,self).__init__(mu, sigma2)
    
class StateSpacePrior(ConditionallyGaussianDistribution):
    def __init__(self, vehicle, par_name_list=[], init_coeffs=None):
        mu = np.zeros(NDOFS)
        Q0 = vehicle.init_noise * np.eye(NDOFS)
        muMap = ConstantMap(len(par_name_list), mu)
        Q0Map = ConstantMap(len(par_name_list), Q0)
        super(StateSpacePrior,self).__init__(
            muMap, sigma=Q0Map, coeffs=init_coeffs)

class DynamicsMap(Map): # Generates matrix Phi
    def __init__(self, vehicle, par_name_list=[]):
        self._vehicle = vehicle
        self._par_name_list = par_name_list
        super(DynamicsMap, self).__init__(
            len(par_name_list), NDOFS**2)
    def evaluate(self, x, *args, **kwargs):
        m = x.shape[0]
        Phi = np.zeros((m, NDOFS, NDOFS))
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            A = self._vehicle.get_system_matrix(**kwargs)
            dt = self._vehicle.dt
            Phi[i,:,:] = scila.expm(dt * A)
        return Phi
    def grad_x(self, x, *args, **kwargs):
        m = x.shape[0]
        gPhi = np.zeros((m, NDOFS, NDOFS, len(self._par_name_list)))
        if len(self._par_name_list) == 0:
            return gPhi
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            A = self._vehicle.get_system_matrix(**kwargs)
            kwargs.update({'par_name_list': self._par_name_list})
            gA = self._vehicle.get_grad_system_matrix(**kwargs)
            dt = self._vehicle.dt
            gPhi[i] = grad_exp(dt*A, dt*gA, startord=50, noit=True)
        return gPhi

class NoiseDynamicsMap(Map): # Generates matrix Q
    def __init__(self, vehicle, par_name_list=[]):
        self._vehicle = vehicle
        self._par_name_list = par_name_list
        super(NoiseDynamicsMap, self).__init__(
            len(par_name_list), NDOFS**2)
    def evaluate(self, x, *args, **kwargs):
        m = x.shape[0]
        Q = np.zeros((m, NDOFS, NDOFS))
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            A = self._vehicle.get_system_matrix(**kwargs)
            G = self._vehicle.get_track_matrix()
            # Dynamic noise
            v = self._vehicle.v
            Ar = self._vehicle.Ar
            dt = self._vehicle.dt
            sigma = np.sqrt( 4 * np.pi**2 * Ar * v**2 * dt)
            W = sigma * np.eye(NDOFS)
            # Pade approximations
            GWG = np.dot(G, np.dot(W, G.T))
            A1 = np.zeros((2*NDOFS,2*NDOFS))
            A1[:NDOFS,:NDOFS] = - A
            A1[:NDOFS,NDOFS:] = GWG
            A1[NDOFS:,NDOFS:] = A.T
            B = scila.expm(dt * A1)
            Phi = B[NDOFS:,NDOFS:].T
            Q[i,:,:] = np.dot(Phi, B[:NDOFS,NDOFS:])
        return Q
    def grad_x(self, x, *args, **kwargs):
        m = x.shape[0]
        dQ = np.zeros((m, NDOFS, NDOFS, len(self._par_name_list)))
        if len(self._par_name_list) == 0:
            return dQ
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            A = self._vehicle.get_system_matrix(**kwargs)
            G = self._vehicle.get_track_matrix()
            kwargs.update({'par_name_list': self._par_name_list})
            dA = self._vehicle.get_grad_system_matrix(**kwargs)
            # Dynamic noise
            v = self._vehicle.v
            Ar = self._vehicle.Ar
            dt = self._vehicle.dt
            sigma = np.sqrt( 4 * np.pi**2 * Ar * v**2 * dt )
            W = sigma * np.eye(NDOFS)
            GWG = np.dot(G, np.dot(W, G.T))
            # Pade approximations Phi, Phi1Q
            A1 = np.zeros((2*NDOFS,2*NDOFS))
            A1[:NDOFS,:NDOFS] = - A
            A1[:NDOFS,NDOFS:] = GWG
            A1[NDOFS:,NDOFS:] = A.T
            B = scila.expm(dt * A1)
            Phi = B[NDOFS:,NDOFS:].T
            Phi1Q = B[:NDOFS,NDOFS:]
            # Pade approximation dPhi
            dPhi = grad_exp(dt*A, dt*dA, startord=50, noit=True)
            # Pade approximation dPhi1Q
            dA1 = np.zeros((2*NDOFS,2*NDOFS,len(self._par_name_list)))
            dA1[:NDOFS,:NDOFS,:] = - dA
            dA1[NDOFS:,NDOFS:,:] = dA.transpose((1,0,2))
            dB = grad_exp(dt*A1, dt*dA1, startord=50, noit=True)
            dPhi1Q = dB[:NDOFS,NDOFS:,:]
            # Compute gradient
            dQ[i] = np.einsum('ijk,jl->ilk', dPhi, Phi1Q) + \
                    np.einsum('ij,jkl->ikl', Phi, dPhi1Q)
        return dQ
        
class StateSpaceTransition(ConditionallyLinearGaussianAR1TransitionDistribution):
    def __init__(self, vehicle, par_name_list=[], init_coeffs=None):
        self.vehicle = vehicle
        c = ConstantMap(len(par_name_list), np.zeros(NDOFS))
        PhiMap = DynamicsMap(vehicle, par_name_list)
        mu = ConstantMap(len(par_name_list), np.zeros(NDOFS))
        QMap = NoiseDynamicsMap(vehicle, par_name_list)
        # Init
        super(StateSpaceTransition,self).__init__(
            c, PhiMap, mu, sigma=QMap, coeffs=init_coeffs)

class ObservationSystemMap(Map): # Generates matrix H
    def __init__(self, vehicle, par_name_list=[]):
        self._vehicle = vehicle
        self._par_name_list = par_name_list
        super(ObservationSystemMap, self).__init__(
            len(par_name_list), len(vehicle.observables)*NDOFS)
    def evaluate(self, x, *args, **kwargs):
        m = x.shape[0]
        H = np.zeros((m, len(self._vehicle.observables), NDOFS))
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            H[i,:,:] = self._vehicle.get_observation_matrix(**kwargs)
        return H
    def grad_x(self, x, *args, **kwargs):
        m = x.shape[0]
        dH = np.zeros((m, len(self._vehicle.observables), NDOFS, len(self._par_name_list)))
        for i in range(m):
            kwargs = {par_name: x[i,p]
                      for p, par_name in enumerate(self._par_name_list)}
            kwargs.update({'par_name_list': self._par_name_list})
            dH[i,:,:,:] = self._vehicle.get_grad_observation_matrix(**kwargs)
        return dH
        
class StateSpaceLogLikelihood(AdditiveConditionallyLinearGaussianLogLikelihood):
    def __init__(self, y, vehicle, par_name_list=[], init_coeffs=np.zeros(0)):
        HMap = ObservationSystemMap(vehicle, par_name_list)
        avars = list(range(len(par_name_list)))
        dy = y.shape[0]
        ###########
        # Original
        R = np.eye(NDOFS)
        R[Yw1dot,Yw1dot] = vehicle.obs_noise['Yw1dot']
        R[Yw2dot,Yw2dot] = vehicle.obs_noise['Yw2dot']
        R[Ybdot,Ybdot]   = vehicle.obs_noise['Ybdot']
        R[Pb,Pb]         = vehicle.obs_noise['Pb']
        R[Ycdot,Ycdot]   = vehicle.obs_noise['Ycdot']
        R = R[np.ix_(vehicle.observables, vehicle.observables)]
        # Prepare maps
        cMap = ConstantMap(len(par_name_list), np.zeros(dy))
        muMap = ConstantMap(len(par_name_list), np.zeros(dy))
        RMap = ConstantMap(len(par_name_list), R)
        super(StateSpaceLogLikelihood,self).__init__(
            y, cMap, HMap, muMap, RMap, coeffs=init_coeffs)

def generate_data(nsteps, vehicle, par_name_list=[], par_val_list=np.zeros(0)):
    prior = StateSpacePrior(vehicle, par_name_list, par_val_list)
    trans = StateSpaceTransition(vehicle, par_name_list, par_val_list)
    ll = StateSpaceLogLikelihood(
        np.zeros(len(vehicle.observables)), vehicle,
        par_name_list, par_val_list) # Just use to get handle on params
    # Generate dynamics
    T = np.zeros(nsteps+1)
    Z = np.zeros((nsteps+1, NDOFS))
    Z[0,:] = prior.rvs(1)[0,:]
    for i in range(nsteps):
        T[i+1] = T[i] + vehicle.dt
        Tin = np.hstack((Z[i,:], par_val_list))[nax,:]
        Z[i+1,:] = trans.T.evaluate(Tin)[0,:] + \
                   trans.pi.rvs(1,par_val_list)[0,:]
    # Generate observations
    Y = []
    for i in range(nsteps+1):
        Y.append( ll.T.evaluate(Z[[i],:])[0,:] + ll.pi.rvs(1)[0,:] )
    return (T,Z,Y)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import numpy as np

from TransportMaps.Distributions.DistributionBase import *
from TransportMaps.Distributions.FrozenDistributions import StandardNormalDistribution

__all__ = ['FactorizedBananaDistribution']

class SquaredAvgNormalDistribution(ConditionalDistribution):
    r"""
    Conditional distribution :math:`x_1 | x_0 \sim \mathcal{N}(x_0^2, 1)`
    """
    def __init__(self):
        super(SquaredAvgNormalDistribution, self).__init__(1,1)
    def log_pdf(self, x, y, params=None, idxs_slice=slice(None,None,None), cache=None):
        return -.5 * (x[:,0] - y[:,0]**2)**2 -.5 * np.log(2 * np.pi)
    def grad_x_log_pdf(
            self, x, y, params=None, idxs_slice=slice(None,None,None), cache=None):
        out = np.zeros((x.shape[0],2))
        out[:,0] = - x[:,0] + y[:,0]**2
        out[:,1] = 2 * x[:,0] * y[:,0] - 2 * y[:,0]**3
        return out
    def hess_x_log_pdf(
            self, x, y, params=None, idxs_slice=slice(None,None,None), cache=None):
        out = np.zeros((x.shape[0],2,2))
        out[:,0,0] = -1.
        out[:,0,1] = 2 * y[:,0]
        out[:,1,0] = out[:,0,1]
        out[:,1,1] = 2 * x[:,0] - 6 * y[:,0]**2
        return out

class FactorizedBananaDistribution(FactorizedDistribution):
    r"""
    Joint distribution :math:`\pi(x_0,x_1)=\pi_1(x_1|x_0)\pi_2(x_0)` defined by

    .. math::

       x_0 \sim \mathcal{N}(0,1) \\
       x_1 | x_0 \sim \mathcal{N}(x_0^2, 1)

    """
    def __init__(self):
        p1 = SquaredAvgNormalDistribution()
        p2 = StandardNormalDistribution(1)
        factors = [(p1, (1,), (0,)),
                   (p2, (0,), ()  )]
        super(FactorizedBananaDistribution,self).__init__(factors)


if __name__ == "__main__":
    import numpy as np
    import matplotlib.pyplot as plt
    d = FactorizedBananaDistribution()
    x = np.linspace(-4,4,30)
    X,Y = np.meshgrid(x,x)
    xx = np.vstack((X.flatten(),Y.flatten())).T
    Z = d.pdf(xx).reshape(X.shape)
    plt.figure()
    plt.contour(X,Y,Z)
    plt.show(False)


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#e
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Transport Map Team
# E-mail: tmteam@mit.edu
# Website: transport-maps.mit.edu
# Support: transport-maps.mit.edu/qa/
#

import numpy as np
import itertools

from TransportMaps.Defaults import Default_IsotropicIntegratedSquaredTriangularTransportMap, Default_IsotropicIntegratedExponentialTriangularTransportMap
from TransportMaps.Distributions.FrozenDistributions import StandardNormalDistribution
from TransportMaps.Distributions.TransportMapDistributions import PushForwardTransportMapDistribution, PullBackTransportMapDistribution
from TransportMaps.Distributions.DistributionBase import DistributionFromSamples
from .GeneralizedPrecision import gen_precision, var_omega
from .RegularizedMap import regularized_map

__all__ = ['SING']

# inputs: data (n x d), polynomial order, node ordering method, tolerance delta
# output: sparse edge set
def SING(data, p_order, ordering, delta, REG=None):
    r""" Identify a sparse edge set of the undirected graph corresponding to the data.

    Args: 
      data: data :math:`{\bf x}_i, i = 1,\dots,n`,
      p_order (int): polynomial order of the transport map representation,
      ordering: scheme used to reorder the variables,
      delta (float): tolerance :math:`\delta`
      REG (dict): regularization dictionary with keys 'type' and 'reweight'

    Returns:
      the generalized precision :math:`\Omega`
    """
    # Print inputs to user
    print("Problem inputs:\n\n dimension = ",data.shape[1],
          "\n number of samples = ",data.shape[0],
          "\n polynomial order = ",p_order,
          "\n ordering type = ", ordering.__class__,
          "\n delta = ",delta,"\n")
    # Initial setup
    # data is (n x d)
    dim = data.shape[1]
    n_samps = data.shape[0]
    nax = np.newaxis
    pi = {}

    # Target density is standard normal
    eta = StandardNormalDistribution(dim)
    # Quadrature type and params
    # 0: Monte Carlo quadrature with n point
    qtype = 0
    qparams = n_samps
    # Gradient information
    # 0: derivative free
    # 1: gradient information
    # 2: Hessian information
    ders = 2
    # Tolerance in optimization
    tol = 1e-5
    # Log stores information from optimization
    log = [ ]

    # All variables are active variables to begin
    active_vars=None
    # Set initial sparsity level
    sparsity = [0]
    sparsityIncreasing = True
    # Number of active variables
    n_active_vars = [(np.power(dim,2) + dim)/2]
    # Create list to store permutations
    perm_list = []
    # Create total_perm vector
    total_perm = np.arange(dim)
    # create counter for iterations
    counter = 0
    while sparsityIncreasing:
    
       # Define base_density from samples
       pi = DistributionFromSamples(data)
    
       # Build the transport map (isotropic for each entry)
       # tm_approx = Default_IsotropicIntegratedSquaredTriangularTransportMap(
       #        dim, p_order, active_vars=active_vars)
       # # Construct density T_\sharp \pi
       # tm_density = PushForwardTransportMapDistribution(tm_approx, pi)
        
       # # SOLVE
       # tm_density.minimize_kl_divergence(eta, qtype=qtype,
       #                                   qparams=qparams,
       #                                   regularization=REG,
       #                                   tol=tol, ders=ders)
       
       tm_approx = regularized_map(eta, pi, data, qtype, qparams, dim, p_order, active_vars, tol, ders, REG)
       
       pb_density = PullBackTransportMapDistribution(tm_approx, eta)

       # Compute generalized precision
       omegaHat = gen_precision(pb_density, data)
       # *********************************************TESTING************************************************************************
       print("omega hat = ", omegaHat)
       
       # Compute variance of generalized precision
       gp_var = var_omega(pb_density, data)
       # *********************************************TESTING************************************************************************
       print("gp var = ", gp_var)
       # Compute tolerance (matrix)
       tau = delta * np.sqrt(gp_var) # set tolerance as square root of variance

       # Save diagonal elements
       omegaHat_diagonal = np.copy(np.diag(omegaHat))
       # Threshold omegaHat
       omegaHat[np.abs(omegaHat) < tau] = 0
       # Put diagonal elements back in
       omegaHat[np.diag_indices_from(omegaHat)] = omegaHat_diagonal
       
       # Reorder variables and data
       perm1 = ordering.perm(omegaHat)
       # Apply to omegaHat
       omegaHat_temp = omegaHat[:,perm1][perm1,:]
       # Check if ordering would flip if applied again (problem for reverse Cholesky)
       perm2 = ordering.perm(omegaHat_temp)

       if (perm2 == perm1).all():
       #if (perm2 == perm1):
           perm_vect = np.arange(dim) # set to identity permutation
       else:
           perm_vect = perm1
       
       # Apply to omegaHat
       omegaHat = omegaHat[:,perm_vect][perm_vect,:]
       # Apply re-ordering to data
       data = data[:, perm_vect]
    
       # Add permutation to perm_list
       perm_list.append(perm_vect)
       # Convolve permutation with total_perm
       total_perm = total_perm[perm_vect]
       inverse_perm = [0] * dim
       for i, p in enumerate(total_perm):
           inverse_perm[p] = i
    
       # Extract lower triangular matrix
       omegaHatLower = np.tril(omegaHat)
       # Count edges
       edge_count = np.count_nonzero(omegaHatLower) - dim
    
       # Variable elimination moving from highest node (dim-1) to node 2 (at most)
       for i in range(dim-1,1,-1):
           non_zero_ind  = np.where(omegaHatLower[i,:i] != 0)[0]
           if len(non_zero_ind) > 1:
               co_parents = list(itertools.combinations(non_zero_ind,2))
               for j in range(len(co_parents)):
                   row_index = max(co_parents[j])
                   col_index = min(co_parents[j])
                   omegaHatLower[row_index, col_index] = 1.0

       # Find list of active_vars
       active_vars = []
       for i in range(dim):
           actives = np.where(omegaHatLower[i,:] != 0)
           active_list = list(set(actives[0]) | set([i]))
           active_list.sort(key=int)
           active_vars.append(active_list)
       
       # Find n_active_vars
       n_active_vars.append(np.sum([len(x) for x in active_vars]))
    
       # Find current sparsity level
       sparsity.append(n_active_vars[0] - n_active_vars[-1])
    
       # Set sparsityIncreasing
       if sparsity[-1] <= sparsity[-2]:
           sparsityIncreasing = False

       # Print statement for SING
       counter = counter + 1
       print('\nSING Iteration: ', counter)
       print('Active variables: ', active_vars, '\n  Note variables may be permuted.')
       print('Number of edges: ', edge_count,' out of ', np.int((dim**2 - dim)/2),' possible')

    # Recovered omega (same variable order as input ordering)
    rec_omega = omegaHat[:,inverse_perm][inverse_perm,:]
    
    print('\nSING has terminated.')
    print('Total permutation used: ',total_perm)
    graph = np.zeros((dim,dim))
    graph[np.nonzero(rec_omega)] = 1
    print('Recovered graph: \n', graph)
    return rec_omega



#!/usr/bin/env python

#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the LGNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# LGNU Lesser General Public License for more details.
#
# You should have received a copy of the LGNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2016 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# E-mail: tmteam@mit.edu
#

import sys, getopt
import os.path
import time, datetime
import dill
import h5py
import numpy as np
import numpy.random as npr
import scipy.stats as scistat
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.collections import PolyCollection
import codecs
import matplotlib.dates as mdates
import matplotlib.cbook as cbook

import TransportMaps as TM
import TransportMaps.Diagnostics as DIAG
import TransportMaps.Densities as DENS

def usage():
    print("postprocess-plotting.py --data=<file_name> " + \
          "--postprocess-data=<fname> " + \
          "[--exchanges=<fname> --events=<fname> " + \
          "--do-smooth --do-unb-smooth --do-filt " + \
          "--do-hyper-smooth --do-hyper-filt --do-hyper-unb-filt --trim-ts=TIMES " + \
          "--do-post-pred --do-unb-post-pred " + \
          "--mcmc-skip=SKIP --mcmc-burnin=BURNIN --ntraj=4 --perc=90,50,20 " + \
          "--do-trim-var-diag --chunk-size=SIZE" + \
          "--store-fig-dir=None]")

def full_usage():
    usage()

def bytespdate2num(fmt, encoding='us-ascii'):
    strconverter = mdates.strpdate2num(fmt)
    def bytesconverter(b):
        s = b.decode(encoding)
        return strconverter(s)
    return bytesconverter

FSIZE = (12,4.5)
LW = 1
TITLES = False
STATE_VAR = "Z_t"
YLIM = [-3.,2.]
    
argv = sys.argv[1:]
DATA = None
POST_DATA = None
EXCHANGES_DATA = None
EVENTS_DATA = None
DO_SMOOTH = False
DO_UNBIASED_SMOOTH = False
DO_FILT = False
DO_HYPER_SMOOTH = False
DO_HYPER_FILT = False
DO_HYPER_UNB_FILT = False
TRIM_TS = None
DO_POST_PRED = False
DO_UNB_POST_PRED = False
DO_TRIM_VAR_DIAG = False
CHUNK_SIZE = None
MCMC_SKIP = 0
MCMC_BURNIN = 0
NTRAJ = 4
PERC_LIST = [90, 50, 20]
PERC_STYLES = [':', '-.', '--']
STORE_FIG_DIR = None
STORE_FIG_FMATS = ['svg', 'pdf']
try:
    opts, args = getopt.getopt(argv,"h",["data=", "postprocess-data=",
                                         "exchanges=", "events=",
                                         "do-smooth", "do-unb-smooth", "do-filt",
                                         "do-hyper-smooth",
                                         "do-hyper-filt", "do-hyper-unb-filt",
                                         "trim-ts=",
                                         "mcmc-skip=", "mcmc-burnin=",
                                         "ntraj=", "perc=",
                                         'do-post-pred', "do-unb-post-pred",
                                         "do-trim-var-diag", "chunk-size=",
                                         "store-fig-dir="])
except getopt.GetoptError:
    full_usage()
    sys.exit(2)
for opt, arg in opts:
    if opt == '-h':
        full_usage()
        sys.exit()
    elif opt == "--data":
        DATA = arg
    elif opt == "--postprocess-data":
        POST_DATA = arg
    elif opt == "--exchanges":
        EXCHANGES_DATA = arg
    elif opt == "--events":
        EVENTS_DATA = arg
    elif opt == "--do-smooth":
        DO_SMOOTH = True
    elif opt == "--do-unb-smooth":
        DO_UNBIASED_SMOOTH = True
    elif opt == "--do-filt":
        DO_FILT = True
    elif opt == "--do-hyper-smooth":
        DO_HYPER_SMOOTH = True
    elif opt == "--do-hyper-filt":
        DO_HYPER_FILT = True
    elif opt == "--do-hyper-unb-filt":
        DO_HYPER_UNB_FILT = True
    elif opt == "--trim-ts":
        TRIM_TS = [ int(t) for t in arg.split(',') ]
    elif opt == "--mcmc-skip":
        MCMC_SKIP = int(arg)
        if MCMC_SKIP < 1:
            raise ValueError("SKIP must be >0 in --mcmc-skip=SKIP")
    elif opt == "--ntraj":
        NTRAJ = int(arg)
    elif opt == "--perc":
        PERC_LIST = [int(p) for p in arg.split(',')]
    elif opt == '--do-post-pred':
        DO_POST_PRED = True
    elif opt == '--do-unb-post-pred':
        DO_UNB_POST_PRED = True
    elif opt == '--do-trim-var-diag':
        DO_TRIM_VAR_DIAG = True
    elif opt == '--chunk-size':
        CHUNK_SIZE = int(arg)
    elif opt == "--store-fig-dir":
        STORE_FIG_DIR = arg

def tstamp_print(msg, *args, **kwargs):
    tstamp = datetime.datetime.fromtimestamp(
        time.time()
    ).strftime('%Y-%m-%d %H:%M:%S')
    print(tstamp + " " + msg, *args, **kwargs)

def filter_tstamp_print(msg, *args, **kwargs):
    if VERBOSE:
        tstamp_print(msg, *args, **kwargs)

def filter_print(*args, **kwargs):
    if VERBOSE:
        print(*args, **kwargs)

def store_figure(fig, fname):
    for fmat in STORE_FIG_FMATS:
        fig.savefig(fname+'.'+fmat, format=fmat, bbox_inches='tight');
        
if None in [DATA, POST_DATA]:
    full_usage()
    tstamp_print("ERROR: Options --data and --postprocess-data must be specified")
    sys.exit(3)

if STORE_FIG_DIR is not None:
    DFILE = DATA.split('/')[-1]
    if not os.path.exists(STORE_FIG_DIR + '/' + DFILE):
        os.makedirs(STORE_FIG_DIR + '/' + DFILE)
    STORE_FIG_DIR += '/' + DFILE +'/'

# Load data
with open(DATA, 'rb') as in_stream:
    stg = dill.load(in_stream)

# Load postprocess data
with open(POST_DATA, 'rb') as istr:
    data = dill.load(istr)

# Load hdf5 file
h5data = h5py.File(POST_DATA + '.hdf5', 'r')

# Load exchanges data
if EXCHANGES_DATA is not None:
    with open(EXCHANGES_DATA,'r') as datafile:
        dates, exchanges, variations = \
            np.loadtxt(datafile, skiprows=2, usecols=[0,1,2], delimiter=',',
                       unpack=True, converters={0: bytespdate2num('%m/%d/%y')})

if EVENTS_DATA is not None:
    with open(EVENTS_DATA,'r') as datafile:
        events_dates = \
            np.loadtxt(datafile, skiprows=1, usecols=[0], delimiter=',',
                       unpack=True, converters={0: bytespdate2num('%m/%d/%Y')})

# Restore data
target_distribution = stg.target_distribution
base_distribution = stg.base_distribution

dim = target_distribution.dim
nsteps = target_distribution.nsteps
nhyper = dim - nsteps
dimfilt = nhyper + 1
if EXCHANGES_DATA is None:
    tvec = np.arange(nsteps)
else:
    tvec = dates[:nsteps]
if EVENTS_DATA is None:
    events_dates = []
else:
    events_dates = [d for d in events_dates if d < tvec[-1]]
obs = target_distribution.observations

if CHUNK_SIZE is None:
    CHUNK_SIZE = nsteps

if target_distribution.is_mu_hyper:
    index_mu = target_distribution.index_mu
if target_distribution.is_sigma_hyper:
    index_sigma = target_distribution.index_sigma
if target_distribution.is_phi_hyper:
    index_phi = target_distribution.index_phi

def nicePlot(ax):
    coloraxes = 'gray'
    ax.spines['right'].set_visible(False) # Remove the right axis boundary
    ax.spines['top'].set_visible(False)  # Remove the top axis boundary
    ax.xaxis.set_ticks_position('bottom') # Set the x-ticks to only the bottom
    ax.yaxis.set_ticks_position('left') # Set the y-ticks to only the left
    ax.spines['bottom'].set_position(('axes',-0.04)) # Offset the bottom scale from the axis
    ax.spines['left'].set_position(('axes',-0.04))  # Offset the left scale from the axis
    ax.spines['left'].set_linewidth(.5)
    ax.spines['bottom'].set_linewidth(.5)
    ax.spines['bottom'].set_color(coloraxes)
    ax.spines['left'].set_color(coloraxes)
    ax.xaxis.label.set_color(coloraxes)
    ax.yaxis.label.set_color(coloraxes)
    ax.tick_params(axis='x', colors=coloraxes)
    ax.tick_params(axis='y', colors=coloraxes)
    ax.yaxis.label.set_size(18)
    ax.xaxis.label.set_size(18)
    ax.title.set_color(coloraxes)
    ax.title.set_size(16)

def plot_mean_percentile_traj(tvec, data1, data2=None, edates=None, ntraj=0, 
                              ylabel=None, title=None, c1='r', c2='k'):
    fig = plt.figure(figsize=FSIZE)
    if EXCHANGES_DATA is not None:
        plt.xlabel("day")
    else:
        plt.xlabel("time")
    if ylabel is not None:
        plt.ylabel(ylabel)
    if TITLES:
        plt.title(title)
    ax = fig.add_subplot(111)
    nicePlot(ax)
    ax.set_xlim(tvec[0], tvec[-1])
    ax.set_ylim(YLIM[0], YLIM[1])
    
    # First data series
    mean1 = np.mean(data1, axis=0)
    ax.plot(tvec, mean1, '-' + c1, linewidth=LW)
    for i, perc in enumerate(PERC_LIST):
        qlow = np.percentile(data1, 50. - perc/2, axis=0)
        qhigh = np.percentile(data1, 50. + perc/2, axis=0)
        ax.plot(tvec, qlow, PERC_STYLES[i] + c1, linewidth=LW)
        ax.plot(tvec, qhigh, PERC_STYLES[i] + c1, linewidth=LW)

    # Trajectories
    for i in range(ntraj):
        ax.plot(tvec, data1[i,:], '-k', linewidth=0.5, alpha=0.6)

    if data2 is not None:
        # Comparing second data series
        mean2 = np.mean(data2, axis=0)
        ax.plot(tvec, mean2, '-' + c2, linewidth=LW)
        for i, perc in enumerate(PERC_LIST):
            qlow = np.percentile(data2, 50. - perc/2, axis=0)
            qhigh = np.percentile(data2, 50. + perc/2, axis=0)
            ax.plot(tvec, qlow, PERC_STYLES[i] + c2, linewidth=LW)
            ax.plot(tvec, qhigh, PERC_STYLES[i] + c2, linewidth=LW)

    if EXCHANGES_DATA is not None:
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%y'))
        fig.autofmt_xdate()
    if EVENTS_DATA is not None:
        for dt in edates:
            ax.plot([dt,dt], YLIM, '--k', linewidth=0.5, alpha=0.6)
            
    return fig

def plot_mean_percentile_traj_3d(data1, ts, data2=None, ntraj=0,
                                 ylabel=None, zlabel=None):
    fig = plt.figure(figsize=FSIZE)
    
    mean_1 = np.mean(data1, axis=0)
    q5_1 = np.percentile(data1, 5, axis=0)
    q25_1 = np.percentile(data1, 25, axis=0)
    q40_1 = np.percentile(data1, 40, axis=0)
    q60_1 = np.percentile(data1, 60, axis=0)
    q75_1 = np.percentile(data1, 75, axis=0)
    q95_1 = np.percentile(data1, 95, axis=0)

    ax = fig.add_subplot(111, projection='3d')
    # Mean and quantiles
    ax.plot(tvec, mean_1, zs=0, zdir='z', c='k')
    verts = []
    verts.append( list(zip(tvec, q40_1)) + list(zip(tvec[::-1], q60_1[::-1])) )
    poly = PolyCollection(verts, facecolor=['red'])
    poly.set_alpha(0.5)
    ax.add_collection3d(poly, zs=[0], zdir='z')
    verts = []
    verts.append( list(zip(tvec, q25_1)) + list(zip(tvec[::-1], q75_1[::-1])) )
    poly = PolyCollection(verts, facecolor=['red'])
    poly.set_alpha(0.3)
    ax.add_collection3d(poly, zs=[0], zdir='z')
    verts = []
    verts.append( list(zip(tvec, q5_1)) + list(zip(tvec[::-1], q95_1[::-1])) )
    poly = PolyCollection(verts, facecolor=['red'])
    poly.set_alpha(0.15)
    ax.add_collection3d(poly, zs=[0], zdir='z')
    miny = np.min(q5_1)
    maxy = np.max(q95_1)
    span = maxy-miny
    miny -= 0.1 * span
    maxy += 0.1 * span
    ax.set_ylim([miny, maxy])
    # Slices
    yy = np.linspace(miny,maxy,100)
    maxz = 0.
    for ii, tt in enumerate(ts):
        # Biased
        ss = data1[:,tt-1]
        kde = scistat.gaussian_kde(ss)
        zz = kde(yy)
        maxz = max(maxz, np.max(zz))
        plt.plot(yy, zz, zs=tvec[tt-1], zdir='x', c='k')
        # Unbiased
        if data2 is not None:
            kde = scistat.gaussian_kde(data2[ii][:])
            zz = kde(yy)
            maxz = max(maxz, np.max(zz))
            plt.plot(yy, zz, '--k', zs=tvec[tt-1], zdir='x') 
    ax.set_zlim([0, maxz*1.05])
    ax.set_xlim([tvec[0], tvec[-1]])
    ax.view_init(elev=20., azim=-120.)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    ax.set_zlabel(zlabel)
    plt.tight_layout()

    if EXCHANGES_DATA is not None:
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%y'))
        fig.autofmt_xdate()
    else:
        ax.set_xlabel('time')
    
    return fig
            
#--1--###### Plot: smoothing/posterior marginals timesteps #############
if DO_SMOOTH:
    print("\nPlotting smoothing/posterior marginals \n")
    tar_samp_smooth = h5data['trim-%d' % nsteps]['quadrature']['approx-target']['0']
    Xt_samp_smooth = tar_samp_smooth[:,nhyper:]
    start = 0
    i = 0
    while start < nsteps:
        stop = min(start + CHUNK_SIZE, nsteps)
        smps = Xt_samp_smooth[:,start:stop]
        tt = tvec[start:stop]
        et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
        fig = plot_mean_percentile_traj(
            tt, smps, ntraj=NTRAJ, edates=et,
            ylabel="$\pi_{" + STATE_VAR + "|Y_{0:N}}$",
            title="Posterior/Smoothing marginals")
        plt.show(False)

        if STORE_FIG_DIR is not None:
            store_figure(fig, STORE_FIG_DIR+'/'+ \
                         'smoothing-marginals-timesteps-ntraj-%d-chunk-%d' % (NTRAJ, i))
        start = stop
        i += 1

if DO_UNBIASED_SMOOTH: # Fast reading from h5 file
    print("\nLoading Metropolis samples \n")
    dset = h5data['trim-%d' % nsteps]\
           ['metropolis-independent-proposal-samples']['skip-%d' % MCMC_SKIP]['x']
    tar_unb_samp_smooth = np.zeros(dset.shape)
    for d in range(dset.shape[1]):
        print("Dimension %d/%d" % (d+1,dset.shape[1]))
        tar_unb_samp_smooth[:,d] = dset[:,d]
        
if DO_SMOOTH and DO_UNBIASED_SMOOTH:
    Xt_unb_samp_smooth = tar_unb_samp_smooth[:,nhyper:]
    start = 0
    i = 0
    while start < nsteps:
        stop = min(start + CHUNK_SIZE, nsteps)
        smps = Xt_samp_smooth[:,start:stop]
        unb_smps = Xt_unb_samp_smooth[:,start:stop]
        tt = tvec[start:stop]
        et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
        fig = plot_mean_percentile_traj(
            tt, smps, unb_smps, edates=et,
            # ylabel= r"$" + STATE_VAR + "$",
            ylabel="$\pi_{" + STATE_VAR + "|Y_{0:N}}$",
            title="Posterior/Smoothing marginals - vs. unbiased")
        plt.show(False)

        if STORE_FIG_DIR is not None:
            store_figure(fig, STORE_FIG_DIR+'/' + \
                         'smoothing-marginals-vs-unbiased-timesteps-chunk-%d' % i)
        start = stop
        i += 1
        
if DO_FILT or DO_HYPER_FILT or DO_HYPER_UNB_FILT:
    fdata = h5data['trim-%d' % nsteps]['filtering']
    nsamps = np.min(
        [ fdata['step-%i/quadrature/0' %i].shape[0]
          for i in range(nsteps) ])
    samp_filt = np.zeros((nsamps,dimfilt,nsteps))
    for n in range(nsteps):
        samp_filt[:,:,n] = fdata['step-%d/quadrature/0' % n][:nsamps,:]

#--2--###### Plot: filtering marginals timesteps #############
if DO_FILT:
    print("Plotting filtering marginals \n")
    Xt_samp_filt = samp_filt[:,-1,:]
    start = 0
    i = 0
    while start < nsteps:
        stop = min(start + CHUNK_SIZE, nsteps)
        smps = Xt_samp_filt[:,start:stop]
        tt = tvec[start:stop]
        et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
        fig = plot_mean_percentile_traj(
            tt, smps, ntraj=NTRAJ, edates=et,
            # ylabel=r"$" + STATE_VAR + "$",
            ylabel=r"$\pi_{" + STATE_VAR + "|Y_{0:t}}$",
            title='Filtering marginals', c1='b')
        plt.show(False)
        
        if STORE_FIG_DIR is not None:
            store_figure(fig, STORE_FIG_DIR+'/filtering-marginals-timesteps-chunk-%d' % i)

        start = stop
        i += 1

#--3--###### Plot: smoothing+filtering marginals timesteps #############
if DO_SMOOTH and DO_FILT:
    print("Plotting smoothing + filtering marginals \n")
    start = 0
    i = 0
    while start < nsteps:
        stop = min(start + CHUNK_SIZE, nsteps)
        smps1 = Xt_samp_smooth[:,start:stop]
        smps2 = Xt_samp_filt[:,start:stop]
        tt = tvec[start:stop]
        et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
        fig = plot_mean_percentile_traj(
            tt, smps1, smps2, edates=et,
            ylabel=r"$" + STATE_VAR + "$",
            title='Smoothing and filtering marginals', c2='b')
        plt.show(False)

        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-filtering-marginals-timesteps-chunk-%d' % i)
        start = stop
        i += 1

#--4--###### Plot: filtering marginals hyper-parameters #############
if DO_HYPER_FILT and nhyper>0:
    hyper_samp_filt = samp_filt[:,:nhyper,:]
    if target_distribution.is_sigma_hyper:
        hyper_samp_filt[:,index_sigma,:] = \
            target_distribution.sigma.evaluate( hyper_samp_filt[:,index_sigma,:] )
    if target_distribution.is_phi_hyper:
        hyper_samp_filt[:,index_phi,:] = \
            target_distribution.phi.evaluate( hyper_samp_filt[:,index_phi,:] )
    
    print("Plotting hyper-parameters filtering marginals \n")
    if target_distribution.is_mu_hyper:
        #Plot filtering marginals for mu
        mu_samp_filt = hyper_samp_filt[:,index_mu,:]
        start = 0
        i = 0
        while start < nsteps:
            stop = min(start + CHUNK_SIZE, nsteps)
            smps = mu_samp_filt[:,start:stop]
            tt = tvec[start:stop]
            et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
            fig = plot_mean_percentile_traj(
                tt, smps, ntraj=0, edates=et,
                ylabel="$\pi_{\mu|Y_{0:t}}$",
                title='Filtering marginals of $\mu$')
            plt.show(False)
            if STORE_FIG_DIR is not None:
                store_figure(
                    fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_mu-chunk-%d' % i)
            start = stop
            i += 1

    if target_distribution.is_sigma_hyper:
        #Plot filtering marginals for sigma
        sigma_samp_filt = hyper_samp_filt[:,index_sigma,:]
        start = 0
        i = 0
        while start < nsteps:
            stop = min(start + CHUNK_SIZE, nsteps)
            smps = sigma_samp_filt[:,start:stop]
            tt = tvec[start:stop]
            et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
            fig = plot_mean_percentile_traj(
                tt, smps, ntraj=0, edates=et,
                ylabel=r"$\pi_{\sigma|Y_{0:t}}$",
                title=r'Filtering marginals of $\sigma$')
            plt.show(False)
            if STORE_FIG_DIR is not None:
                store_figure(
                    fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_sigma-chunk-%d' % i)
            start = stop
            i += 1

    if target_distribution.is_phi_hyper:
        #Plot filtering marginals for phi
        phi_samp_filt = hyper_samp_filt[:,index_phi,:]
        start = 0
        i = 0
        while start < nsteps:
            stop = min(start + CHUNK_SIZE, nsteps)
            smps = phi_samp_filt[:,start:stop]
            tt = tvec[start:stop]
            et = [t for t in events_dates if tt[0] <= t <= tt[-1]]
            fig = plot_mean_percentile_traj(
                tt, smps, ntraj=0, edates=et,
                ylabel=r"$\pi_{\phi|Y_{0:t}}$",
                title=r'Filtering marginals of $\sigma$')
            plt.show(False)
            if STORE_FIG_DIR is not None:
                store_figure(
                    fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_phi-chunk-%d' % i)
            start = stop
            i += 1
        
if DO_HYPER_UNB_FILT and nhyper>0:
    tar_unb_samp_smooth_trim = []
    for t in TRIM_TS:
        h5root = h5data['trim-%d' % t]
        tar_unb_samp_smooth_trim.append(
            h5root['metropolis-independent-proposal-samples/skip-%d/x/' % MCMC_SKIP]
        )
    print("Plotting hyper-parameters filtering marginals vs. unbiased \n")
    if target_distribution.is_mu_hyper:
        # Plot filtering marginals for mu (3D)
        index_mu = target_distribution.index_mu
        mu_samp_filt = hyper_samp_filt[:,index_mu,:]
        mu_unb_samp_smooth = [ samp[:,index_mu] for samp in tar_unb_samp_smooth_trim ]
        ylabel = "" # r"$\mu\vert Y_{0:t}$"
        zlabel = r"$\pi_{\mu\vert Y_{0:t}}$"
        fig = plot_mean_percentile_traj_3d(
            mu_samp_filt, TRIM_TS, mu_unb_samp_smooth,
            ylabel=ylabel, zlabel=zlabel)
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_mu-3d')
            
    if target_distribution.is_sigma_hyper:
        # Plot filtering marginals for mu (3D)
        index_sigma = target_distribution.index_sigma
        sigma_samp_filt = hyper_samp_filt[:,index_sigma,:]
        sigma_unb_samp_smooth = [
            target_distribution.sigma.evaluate( samp[:,index_sigma] )
            for samp in tar_unb_samp_smooth_trim ]
        ylabel = "" # r"$\sigma\vert Y_{0:t}$"
        zlabel = r"$\pi_{\sigma\vert Y_{0:t}}$"
        fig = plot_mean_percentile_traj_3d(
            sigma_samp_filt, TRIM_TS, sigma_unb_samp_smooth,
            ylabel=ylabel, zlabel=zlabel)
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_sigma-3d')
            
    if target_distribution.is_phi_hyper:
        # Plot filtering marginals for mu (3D)
        index_phi = target_distribution.index_phi
        phi_samp_filt = hyper_samp_filt[:,index_phi,:]
        phi_unb_samp_smooth = [
            target_distribution.phi.evaluate( samp[:,index_phi] )
            for samp in tar_unb_samp_smooth_trim ]
        ylabel = "" # r"$\phi\vert Y_{0:t}$"
        zlabel = r"$\pi_{\phi\vert Y_{0:t}}$"
        fig = plot_mean_percentile_traj_3d(
            phi_samp_filt, TRIM_TS, phi_unb_samp_smooth,
            ylabel=ylabel, zlabel=zlabel)
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/filtering-marginals-timesteps_phi-3d')
            
#--4--###### Plot: smoothing/posterior marginals hyper-parameters #############
if DO_HYPER_SMOOTH and nhyper>0:
    tar_samp_smooth = h5data['trim-%d' % nsteps]['quadrature']['approx-target']['0']
    print("Plotting hyper-parameters smoothing marginals \n")
    if target_distribution.is_mu_hyper:
        mu_samp_smooth = tar_samp_smooth[:,index_mu]
        mu_min = np.min(mu_samp_smooth)
        mu_min = mu_min - .1*np.abs(mu_min)
        mu_max = np.max(mu_samp_smooth)
        mu_max = mu_max + .1*np.abs(mu_max)
        mu_kde = scistat.gaussian_kde(mu_samp_smooth)
        mu_xx = np.linspace(mu_min, mu_max, 10000)
        pdf_mu_xx = mu_kde(mu_xx)
        fig = plt.figure()
        plt.xlabel("$\mu$")
        plt.ylabel("$\pi_{\mu|Y_{0:N}}$")
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\mu$')
        plt.xlim(mu_min,mu_max)
        plt.plot(mu_xx, pdf_mu_xx, '-k',linewidth=1.0)
        plt.show(False)

        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-mu') 

    if target_distribution.is_sigma_hyper:
        Xsigma_samp_smooth =  tar_samp_smooth[:,index_sigma]
        # Do the kde on the reference variable and then push forward through F_sigma.
        Xsigma_kde = scistat.gaussian_kde(Xsigma_samp_smooth) 
        Xsigma_min = np.min(Xsigma_samp_smooth)
        Xsigma_min = Xsigma_min - .1*np.abs(Xsigma_min)
        Xsigma_max = np.max(Xsigma_samp_smooth)
        Xsigma_max = Xsigma_max + .1*np.abs(Xsigma_max)
        Xsigma_xx = np.linspace(Xsigma_min, Xsigma_max, 10000)
        FXsigma_xx = target_distribution.sigma.evaluate( Xsigma_xx )
        grad_FXsigma_xx = target_distribution.sigma.grad_x( Xsigma_xx )
        pdf_sigma_FXsigma = Xsigma_kde(Xsigma_xx)/grad_FXsigma_xx
        fig = plt.figure()
        plt.xlabel("$\sigma$")
        plt.ylabel("$\pi_{\sigma|Y_{0:N}}$")
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\sigma$')
        sigma_max = np.max(FXsigma_xx)
        ax.set_xlim(0.,sigma_max)
        ax.plot(FXsigma_xx, pdf_sigma_FXsigma, '-k',linewidth=1.0)
        plt.show(False)

        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-sigma') 

    if target_distribution.is_phi_hyper:
        Xphi_samp_smooth =  target_distribution.phi.evaluate(
            tar_samp_smooth[:,index_phi])
        Xphi_kde = scistat.gaussian_kde(Xphi_samp_smooth)
        Xphi_min = np.min(Xphi_samp_smooth)
        Xphi_max = np.max(Xphi_samp_smooth)
        Xphi_xx = np.linspace(Xphi_min, Xphi_max, 1000)
        pdf_phi = Xphi_kde(Xphi_xx)
        fig = plt.figure()
        plt.ylabel("$\pi_{\phi|Y_{0:N}}$")
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\phi$')
        plt.plot(Xphi_xx, pdf_phi, '-k',linewidth=1.0, label='Transport Map')
        plt.legend(loc='best')
        plt.show(False)
        
        # Xphi_samp_smooth =  tar_samp_smooth[:,index_phi]
        # Xphi_kde = scistat.gaussian_kde(Xphi_samp_smooth)
        # #Do the kde on the reference variable and then push forward through F_phi.
        # Xphi_min = np.min(Xphi_samp_smooth)
        # Xphi_min = Xphi_min # - .1*np.abs(Xphi_min)
        # Xphi_max = np.max(Xphi_samp_smooth)
        # Xphi_max = Xphi_max # + .1*np.abs(Xphi_max)
        # Xphi_xx = np.linspace(Xphi_min, Xphi_max, 10000)
        # FXphi_xx = target_distribution.phi.evaluate( Xphi_xx )
        # grad_FXphi_xx = target_distribution.phi.grad_x( Xphi_xx )
        # pdf_phi_FXphi = Xphi_kde(Xphi_xx)/grad_FXphi_xx
        # fig = plt.figure()
        # plt.xlabel("$\phi$")
        # plt.ylabel("$\pi_{\phi|Y_{0:N}}$")
        # if TITLES:
        #     plt.title('Posterior/smoothing marginals of $\phi$')
        # phi_min = np.min(FXphi_xx)
        # plt.xlim(phi_min,1.)
        # plt.plot(FXphi_xx, pdf_phi_FXphi, '-k',linewidth=1.0)
        # plt.show(False)
        # #phi_samp_smooth = target_distribution.phi.evaluate( Xphi_samp_smooth )

        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-phi')

    print("Plotting posterior/smoothing marginals \n")
    hyper_samp_smooth = tar_samp_smooth[:,:nhyper].copy()
    if target_distribution.is_sigma_hyper:
        hyper_samp_smooth[:,index_sigma] = target_distribution.sigma.evaluate(
            hyper_samp_smooth[:,index_sigma] )
    if target_distribution.is_phi_hyper:
        hyper_samp_smooth[:,index_phi] = target_distribution.phi.evaluate(
            hyper_samp_smooth[:,index_phi] )
    (fig, handles,_,_,_) = DIAG.plotAlignedMarginals(
        hyper_samp_smooth, show_axis=True )
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/smoothing-marginals-hyper')

if DO_HYPER_SMOOTH and DO_UNBIASED_SMOOTH and nhyper > 0:
    print("Plotting hyper-parameters unbiased smoothing marginals \n")
    hyper_unb_samp_smooth = tar_unb_samp_smooth[:,:nhyper].copy()
    if target_distribution.is_sigma_hyper:
        hyper_unb_samp_smooth[:,index_sigma] = target_distribution.sigma.evaluate(
            hyper_unb_samp_smooth[:,index_sigma] )
    if target_distribution.is_phi_hyper:
        hyper_unb_samp_smooth[:,index_phi] = target_distribution.phi.evaluate(
            hyper_unb_samp_smooth[:,index_phi] )
    # (fig, handles,_,_,_) = DIAG.plotAlignedMarginals(
    #     hyper_unb_samp_smooth, show_axis=True )
    (fig, handles,_,_,_) = DIAG.plotAlignedMarginals(
        hyper_samp_smooth, hyper_unb_samp_smooth, show_axis=True )
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/unbiased-smoothing-marginals-hyper')

    
    print("Plotting hyper-parameters smoothing marginals vs. unbiased \n")
    if target_distribution.is_mu_hyper:
        mu_samp_smooth = tar_samp_smooth[:,index_mu]
        mu_unb_samp_smooth = tar_unb_samp_smooth[:,index_mu]
        mu_min = min(np.min(mu_samp_smooth), np.min(mu_unb_samp_smooth))
        mu_min = mu_min - .1*np.abs(mu_min)
        mu_max = max(np.max(mu_samp_smooth), np.max(mu_unb_samp_smooth))
        mu_max = mu_max + .1*np.abs(mu_max)
        mu_kde = scistat.gaussian_kde(mu_samp_smooth)
        mu_unb_kde = scistat.gaussian_kde(mu_unb_samp_smooth)
        mu_xx = np.linspace(mu_min, mu_max, 10000)
        pdf_mu_xx = mu_kde(mu_xx)
        pdf_mu_unb_xx = mu_unb_kde(mu_xx)
        fig = plt.figure()
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\mu$')
        plt.ylabel("$\pi_{\mu|Y_{0:N}}$")
        plt.xlim(mu_min,mu_max)
        plt.plot(mu_xx, pdf_mu_xx, '-k',linewidth=1.0, label='Transport map')
        plt.plot(mu_xx, pdf_mu_unb_xx, '--k',linewidth=1.0, label='MCMC')
        plt.legend(loc='best')
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-mu-vs-unb') 

    if target_distribution.is_sigma_hyper:
        Xsigma_samp_smooth =  tar_samp_smooth[:,index_sigma]
        Xsigma_unb_samp_smooth = tar_unb_samp_smooth[:,index_sigma]
        # Do the kde on the reference variable and then push forward through F_sigma.
        Xsigma_kde = scistat.gaussian_kde(Xsigma_samp_smooth)
        Xsigma_unb_kde = scistat.gaussian_kde(Xsigma_unb_samp_smooth) 
        Xsigma_min = min(np.min(Xsigma_samp_smooth), np.min(Xsigma_unb_samp_smooth))
        Xsigma_min = Xsigma_min - .1*np.abs(Xsigma_min)
        Xsigma_max = max(np.max(Xsigma_samp_smooth), np.max(Xsigma_unb_samp_smooth))
        Xsigma_max = Xsigma_max + .1*np.abs(Xsigma_max)
        Xsigma_xx = np.linspace(Xsigma_min, Xsigma_max, 10000)
        FXsigma_xx = target_distribution.sigma.evaluate( Xsigma_xx )
        grad_FXsigma_xx = target_distribution.sigma.grad_x( Xsigma_xx )
        pdf_sigma_FXsigma = Xsigma_kde(Xsigma_xx)/grad_FXsigma_xx
        pdf_unb_sigma_FXsigma = Xsigma_unb_kde(Xsigma_xx)/grad_FXsigma_xx
        fig = plt.figure()
        plt.ylabel("$\pi_{\sigma|Y_{0:N}}$")
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\sigma$')
        sigma_max = np.max(FXsigma_xx)
        plt.xlim(0.,sigma_max)
        plt.plot(FXsigma_xx, pdf_sigma_FXsigma, '-k',linewidth=1.0, label='Transport Map')
        plt.plot(FXsigma_xx, pdf_unb_sigma_FXsigma, '--k',
                 linewidth=1.0, label='MCMC')
        plt.legend(loc='best')
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-sigma-vs-unb') 

    if target_distribution.is_phi_hyper:
        Xphi_samp_smooth =  target_distribution.phi.evaluate(
            tar_samp_smooth[:,index_phi])
        Xphi_unb_samp_smooth =  target_distribution.phi.evaluate(
            tar_unb_samp_smooth[:,index_phi])
        Xphi_kde = scistat.gaussian_kde(Xphi_samp_smooth)
        Xphi_unb_kde = scistat.gaussian_kde(Xphi_unb_samp_smooth)
        Xphi_min = min(np.min(Xphi_samp_smooth), np.min(Xphi_unb_samp_smooth))
        Xphi_min = Xphi_min # - .1*np.abs(Xphi_min)
        Xphi_max = max(np.max(Xphi_samp_smooth), np.max(Xphi_unb_samp_smooth))
        Xphi_max = Xphi_max # + .1*np.abs(Xphi_max)
        Xphi_xx = np.linspace(Xphi_min, Xphi_max, 1000)
        pdf_phi = Xphi_kde(Xphi_xx)
        pdf_unb_phi = Xphi_unb_kde(Xphi_xx)
        fig = plt.figure()
        plt.ylabel("$\pi_{\phi|Y_{0:N}}$")
        if TITLES:
            plt.title('Posterior/smoothing marginals of $\phi$')
        plt.plot(Xphi_xx, pdf_phi, '-k',linewidth=1.0, label='Transport Map')
        plt.plot(Xphi_xx, pdf_unb_phi, '--k',linewidth=1.0, label='MCMC')
        plt.legend(loc='best')
        plt.show(False)
        if STORE_FIG_DIR is not None:
            store_figure(
                fig, STORE_FIG_DIR+ '/smoothing-marginals-phi-vs-unb')
        
        # Xphi_samp_smooth =  tar_samp_smooth[:,index_phi]
        # Xphi_unb_samp_smooth =  tar_unb_samp_smooth[:,index_phi]
        # #Do the kde on the reference variable and then push forward through F_phi.
        # Xphi_kde = scistat.gaussian_kde(Xphi_samp_smooth)
        # Xphi_unb_kde = scistat.gaussian_kde(Xphi_unb_samp_smooth) 
        # Xphi_min = min(np.min(Xphi_samp_smooth), np.min(Xphi_unb_samp_smooth))
        # Xphi_min = Xphi_min # - .1*np.abs(Xphi_min)
        # Xphi_max = max(np.max(Xphi_samp_smooth), np.max(Xphi_unb_samp_smooth))
        # Xphi_max = Xphi_max # + .1*np.abs(Xphi_max)
        # Xphi_xx = np.linspace(Xphi_min, Xphi_max, 1000)
        # FXphi_xx = target_distribution.phi.evaluate( Xphi_xx )
        # grad_FXphi_xx = target_distribution.phi.grad_x( Xphi_xx )
        # pdf_phi_FXphi = Xphi_kde(Xphi_xx)/grad_FXphi_xx
        # pdf_unb_phi_FXphi = Xphi_unb_kde(Xphi_xx)/grad_FXphi_xx
        # fig = plt.figure()
        # plt.ylabel("$\pi_{\phi|Y_{0:N}}$")
        # if TITLES:
        #     plt.title('Posterior/smoothing marginals of $\phi$')
        # phi_min = np.min(FXphi_xx)
        # plt.plot(FXphi_xx, pdf_phi_FXphi, '-k',linewidth=1.0, label='Transport Map')
        # plt.plot(FXphi_xx, pdf_unb_phi_FXphi, '--k',linewidth=1.0, label='MCMC')
        # plt.legend(loc='best')
        # plt.show(False)
        # #phi_samp_smooth = target_distribution.phi.evaluate( Xphi_samp_smooth )
        # if STORE_FIG_DIR is not None:
        #     store_figure(
        #         fig, STORE_FIG_DIR+ '/smoothing-marginals-phi-vs-unb')

#--7--###### Plot: posterior data predictive #############
def plot_post_pred(obs, data1, data2=None, edates=None, ylabel=None, title=None):
    fig = plt.figure(figsize=FSIZE)
    ax = fig.add_subplot(111)
    nicePlot(ax)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if TITLES:
        plt.title(title)
    
    q5_1 = np.percentile(data1, 5, axis=0)
    q25_1 = np.percentile(data1, 25, axis=0)
    q40_1 = np.percentile(data1, 40, axis=0)
    q60_1 = np.percentile(data1, 60, axis=0)
    q75_1 = np.percentile(data1, 75, axis=0)
    q95_1 = np.percentile(data1, 95, axis=0)
    ax.set_xlim(tvec[0],tvec[-1])
    ax.scatter(tvec,obs,s=0.5,c='k',label="data")
    ax.fill_between(tvec, q40_1, q60_1, color="r", alpha=0.6)
    ax.fill_between(tvec, q25_1, q75_1, color="r", alpha=0.3)
    ax.fill_between(tvec, q5_1, q95_1, color="r", alpha=0.1)

    if data2 is not None:
        q5_2 = np.percentile(data2, 5, axis=0)
        q25_2 = np.percentile(data2, 25, axis=0)
        q40_2 = np.percentile(data2, 40, axis=0)
        q60_2 = np.percentile(data2, 60, axis=0)
        q75_2 = np.percentile(data2, 75, axis=0)
        q95_2 = np.percentile(data2, 95, axis=0)
        ax.fill_between(tvec, q40_2, q60_2, color="b", alpha=0.6)
        ax.fill_between(tvec, q25_2, q75_2, color="b", alpha=0.3)
        ax.fill_between(tvec, q5_2, q95_2, color="b", alpha=0.1)

    if EXCHANGES_DATA is not None:
        ax.set_xlabel("day")
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%m/%d/%y'))
        fig.autofmt_xdate()
    else:
        ax.set_xlabel("time")
    if EVENTS_DATA is not None:
        for dt in edates:
            ax.plot([dt,dt], YLIM, '--k')
    
    ax.legend(loc='best')
    return fig

def plot_qq_post_pred(obs, data, ylabel=None, xlabel=None, title=None):
    nprc = 21
    prclst = np.linspace(0,100,nprc)

    prc = np.zeros((nprc,data.shape[1]))
    for n, p in enumerate(prclst):
        prc[n,:] = np.percentile(data, p, axis=0)

    nobs_in_prc = np.zeros(nprc)
    for n, p in enumerate(prclst):
        nobs_in_prc[n] = sum( obs < prc[n,:] )
        
    # qq-plot and identity tendency line
    fig = plt.figure()
    plt.scatter(prclst/100.,
                nobs_in_prc/float(len(obs)), s=5, c='k')
    plt.plot([0,1],[0,1],'r')
    if ylabel is not None:
        plt.ylabel(ylabel)
    if xlabel is not None:
        plt.xlabel(xlabel)
    if TITLES:
        plt.title(title)

    # # Perform PCA
    # P = np.vstack((p1,p2)).T
    # mP = np.mean(P,axis=0)
    # plt.scatter(mP[0],mP[1], s=15, c='k', marker='s')
    # X = P - mP
    # V = np.dot(X.T,X)
    # w,v = npla.eig(V)
    # srtidxs = np.argsort(w)[::-1]
    # w = w[srtidxs]
    # v = v[:,srtidxs]
    # v0 = v[:,0]
    # v1 = v[:,1]
    # yy = mP[1] + v0[1]/v0[0] * (xx - mP[0])
    # prjv1 = np.outer(v1,v1)
    # res = np.dot(prjv1, X.T).T
    # stdres = np.std(npla.norm(res,axis=1))
    # xy = np.vstack( (xx,yy) )
    # xyp = xy + 1.96 * stdres * v1[:,np.newaxis]
    # xym = xy - 1.96 * stdres * v1[:,np.newaxis]
    # # Plot PCA and 95% confidence interval
    # plt.plot(xx, yy,'k')
    # plt.plot(xyp[0,:],xyp[1,:],'--k')
    # plt.plot(xym[0,:],xym[1,:],'--k')
    return fig
    
if DO_POST_PRED:
    tar_samp_smooth = h5data['trim-%d' % nsteps]['quadrature']['approx-target']['0']
    Xt_samp_smooth = tar_samp_smooth[:,nhyper:]
    noise_samples = npr.randn(Xt_samp_smooth.shape[0],Xt_samp_smooth.shape[1])
    Ysmooth_samples = noise_samples*np.exp(.5*Xt_samp_smooth)
    print("Plotting posterior predictive \n")
    fig = plot_post_pred(
        obs, Ysmooth_samples, ylabel="$\pi_{Y_y|Y_{0:N}}$",
        title="Data Posterior/Smoothing marginals (posterior predictive)")
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/posterior-data-predictive')

    fig = plot_qq_post_pred(obs, Ysmooth_samples,
                            xlabel='quantile', ylabel='observations',
                            title='qq-plot')
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR + '/qq-plot-posterior-data-predictive')

if DO_UNB_POST_PRED:
    Xt_samp_smooth = tar_unb_samp_smooth[:,nhyper:]
    noise_samples = npr.randn(Xt_samp_smooth.shape[0],Xt_samp_smooth.shape[1])
    unb_Ysmooth_samples = noise_samples*np.exp(.5*Xt_samp_smooth)
    print("Plotting unbiased posterior predictive \n")
    fig = plot_post_pred(
        obs, unb_Ysmooth_samples, ylabel="$\pi_{Y_y|Y_{0:N}}$",
        title="Data Posterior/Smoothing marginals (posterior predictive)")
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/posterior-data-predictive-unbiased')

    fig = plot_qq_post_pred(obs, unb_Ysmooth_samples,
                            xlabel='quantile', ylabel='observations',
                            title='qq-plot')
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR + '/unb-qq-plot-posterior-data-predictive')

if DO_POST_PRED and DO_UNB_POST_PRED:
    print("Plotting posterior predictive vs. unbiased posterior predictive\n")
    fig = plot_post_pred(
        obs, Ysmooth_samples, unb_Ysmooth_samples,
        ylabel="$\pi_{Y_y|Y_{0:N}}$",
        title="Data Posterior/Smoothing marginals (posterior predictive)")
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/posterior-data-predictive-vs-unbiased')

if DO_TRIM_VAR_DIAG:
    var_diag_list = []
    for t in TRIM_TS:
        h5root = h5data['trim-%d' % t]
        vals = h5root['vals_var_diag']['exact-base']['0']
        vals_d1 = vals['vals_d1'][:]
        vals_d2 = vals['vals_d2'][:]
        N = len(vals_d1)
        w = np.ones(N)/float(N)
        var_diag_list.append(
            DIAG.variance_approx_kl(
                None, None, vals_d1=vals_d1, vals_d2=vals_d2, w=w) )
    fig = plt.figure()
    ax = fig.add_subplot(111)
    nicePlot(ax)
    ax.semilogy(TRIM_TS, var_diag_list, '-ok')
    ax.grid(True)
    ax.set_ylabel(r"$\mathcal{D}_{\rm KL}(\rho \Vert T^\sharp \pi)$")
    ax.set_xlabel("time")
    plt.show(False)
    if STORE_FIG_DIR is not None:
        store_figure(
            fig, STORE_FIG_DIR+ '/var-diag-evolution')


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Authors: Daniele Bigoni
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

try:
    from lxml import etree
    LXML_SUPPORT = True
except ImportError:
    import xml.etree.ElementTree as etree
    LXML_SUPPORT = False

__all__ = ['load_xml', 'from_xml_element',
           'store_xml', 'to_xml_element',
           'id_parser', 'vars_parser',
           'num_parser', 'midx_parser']

def load_xml(fname):
    import os
    if LXML_SUPPORT:
        tmxml_dir = os.path.dirname(os.path.realpath(__file__))
        schema_path = tmxml_dir + '/schema/map.xsd'
        schema_tree = etree.parse(schema_path)
        schema_root = schema_tree.getroot()
        schema = etree.XMLSchema(schema_root)
        parser = etree.XMLParser(schema=schema)
        tree = etree.parse(fname, parser)
        root = tree.getroot()
    else:
        tree = etree.parse(fname)
        root = tree.getroot()

    return from_xml_element(root)

def from_xml_element(node):
    from TransportMaps import XML_NAMESPACE
    from TransportMaps import Maps as MAPS
    if node.tag == XML_NAMESPACE + 'trimap':
        return MAPS.TriangularTransportMap.from_xml_element(node)

def store_xml(obj, fname):
    raise NotImplementedError("Coming soon")
    
def to_xml_element(obj):
    raise NotImplementedError("Coming soon")
    
def id_parser(id_field, dim):
    if id_field == '*':
        ncomp_list = range(dim)
    elif ':' in id_field:
        sp = id_field.split(':')
        sp_int = [num_parser(s, d=dim-1) for s in sp]
        if id_field[0] == ':':
            ncomp_list = range(sp_int[1])
        elif id_field[-1] == ':':
            ncomp_list = range(sp_int[0],dim)
        else:
            ncomp_list = range(sp_int[0],sp_int[1])
    else:
        ncomp_list = [ num_parser(id_field, dim) ]
    return ncomp_list

def vars_parser(vars_field, d):
    if vars_field == '*':
        vars_list = range(d+1)
    elif ':' in vars_field:
        sp = vars_field.split(':')
        sp_int = [ num_parser(s, d=d) for s in sp ]
        if vars_field[0] == ':':
            vars_list = range(sp_int[1])
        elif vars_field[-1] == ':':
            start = max(0, sp_int[0])
            vars_list = range(start, d+1)  # Changed to d+1
        else:
            start = max(0, sp_int[0])
            vars_list = range(start, sp_int[1])
    else: # Regular string
        vars_list = [ num_parser(vars_field, d) ]
    return vars_list

def num_parser(nstr, d):
    if nstr == '':
        out = None
    elif nstr[0] == 'd': # Format d-x or d+x
        sp_plus = nstr.split('+')
        sp = [ s.split('-') for s in sp_plus ]
        out = 0
        for s0 in sp:
            if s0[0] == 'd':
                sub = d
            else:
                sub = int(s0[0])
            for s1 in s0[1:]:
                sub -= int(s1)
            out += sub
    else:
        out = int(nstr)
    return out

def midx_parser(midx_str):
    midx = tuple([ int(s) for s in midx_str.split(',') ])
    return midx
    


#
# This file is part of TransportMaps.
#
# TransportMaps is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TransportMaps is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with TransportMaps.  If not, see <http://www.gnu.org/licenses/>.
#
# Transport Maps Library
# Copyright (C) 2015-2018 Massachusetts Institute of Technology
# Uncertainty Quantification group
# Department of Aeronautics and Astronautics
#
# Author: Transport Map Team
# Website: transportmaps.mit.edu
# Support: transportmaps.mit.edu/qa/
#

import sys
import os
import os.path
import dill
import numpy as np

from . import AvailableOptions as AO
from .PostprocessBase import Postprocess

from TransportMaps.External import H5PY_SUPPORT
import TransportMaps.Diagnostics as DIAG
import TransportMaps.Samplers as SAMP
import TransportMaps.Distributions as DIST

if H5PY_SUPPORT:
    import h5py

__all__ = ['SequentialPostprocess']

class SequentialPostprocess(Postprocess):
    def usage(self):
        usage_str = """
Usage: tmap-sequential-postprocess [-h -v -I]
  --input=INPUT --output=OUTPUT
  [--trim=NSTEPS
   --store-fig-dir=DIR --store-fig-fmats=FMATS
   --extra-tit=TITLE --no-plotting
   --trim=NSTEPS
   --sequential-var-diag
   --sequential-reg-diag
   --filtering-conditionals
     --filt-alc-n-points-x-ax=N --filt-alc-n-tri-plots=N
     --filt-alc-anchor=LIST --filt-alc-range=LIST
   --filtering-marginals
     --filt-alm-n-points=N --filt-alm-n-tri-plots=N
   --filtering-quadrature
     --filt-quad-qtype=QTYPE
     --filt-quad-qnum=QNUM
   --log=LOG --batch=BATCH --nprocs=NPROCS]
"""
        print(usage_str)

    def description(self):
        docs_distributions_str = \
            AO.print_avail_options(AO.AVAIL_DISTRIBUTIONS,
                                      '                        ', False)
        docs_log_str = \
            '  --log=LOG               log level (default=30). Uses package logging.\n' + \
            AO.print_avail_options(AO.AVAIL_LOGGING,'                          ')
        docs_str = """DESCRIPTION
Given a file (--input) storing the transport map pushing forward a base distribution
to a sequential Hidden Markov target distribution, 
provides a number of postrprocessing routines.
All files involved are stored and loaded using the python package dill and
an extra file OUTPUT.hdf5 is created to store big datasets in the hdf5 format.
In the following default values are shown in brackets.

All the options available for tmaps-postprocess are also available here.

OPTIONS - input/output:
  --input=INPUT         path to the file containing the target distribution,
                          the base distribution and the transport map pushing forward
                          the base to the target.
  --output=OUTPUT       path to the file storing all postprocess data.
                          The additional file OUTPUT.hdf5 will be used to store
                          the more memory consuming data.
  --trim=NSTEPS         trim the solution to NSTEPS and perform analysis
  --store-fig-dir=DIR   path to the directory where to store the figures.
  --store-fig-fmats=FMATS  figure formats - see matplotlib for supported formats (svg)
  --extra-tit=TITLE     additional title for the figures' file names.
  --no-plotting         do not plot figures, but only store their data.
                          (requires --output or --store-fig-dir)
  --trim=NSTEPS         trims the results to NSTEPS and run diagnostics on this distribution
OPTIONS - Diagnostics:
  --sequential-var-diag plot value of variance diagnostic for the sequence of maps
  --sequential-reg-diag plot value of all regression residuals
  --filtering-conditionals  plot aligned slices of the filtering distribution:
                        Optional arguments:
    --filt-alc-n-points-x-ax=N  number of discretization points per axis (30)
    --filt-alc-n-tri-plots=N    number of subplots (0)
    --filt-alc-anchor=LIST      list of floats "f1,f2,f3..." for the anchor point (0)
    --filt-alc-range=LIST       list of two floats "f1,f2" for the range (-5,5)
OPTIONS - Sampling:
  --filtering-marginals  plot aligned marginals of the filtering distribution:
                        Optional arguments:
    --filt-alm-n-points=N     number of samples to be used for the kernel density estimation
    --filt-alm-n-tri-plots=N  number of subplots (0)
  --filtering-quadrature  generate quadrature of the filtering distribution:
                        Optional arguments:
    --filt-quad-qtype=QTYPE  generate quadrature of type QTYPE (0)
    --filt-quad-qnum=QNUM    level of the quadrature (int or list)
OPTIONS - Computation:
""" + docs_log_str + """  --nprocs=NPROCS         number of processors to be used (default=1)
  --batch=BATCH           list of batch sizes for function evaluation, gradient
                          evaluation and Hessian evaluation
OPTIONS - other:
  -v                      verbose output (not affecting --log)
  -I                      enter interactive mode after finishing
  -h                      print this help
"""
        print(docs_str)

    @property
    def long_options(self):
        return super(SequentialPostprocess, self).long_options + \
            [
                "trim=",
                # Sequential diagnostics
                "sequential-var-diag", "sequential-reg-diag",
                # Aligned conditionals
                "filtering-conditionals",
                "filt-alc-n-points-x-ax=", "filt-alc-n-tri-plots=", 
                "filt-alc-anchor=", "filt-alc-range=",
                # Aligned marginals
                "filtering-marginals",
                "filt-alm-n-points=", "filt-alm-n-tri-plots=",
                # Quadrature
                "filtering-quadrature", 
                "filt-quad-qtype=", "filt-quad-qnum=",
            ]

    def load_opts(self, opts):
        super(SequentialPostprocess, self).load_opts(opts)
        for opt, arg in opts:
            if opt == "--trim":
                self.TRIM = int(arg)
            # Sequential diagnostics
            elif opt == "--sequential-var-diag":
                self.SEQUENTIAL_VAR_DIAG = True
            elif opt == "--sequential-reg-diag":
                self.SEQUENTIAL_REG_DIAG = True

            # Aligned conditionals
            elif opt in ("--filtering-conditionals"):
                self.FLT_ALIGNED_CONDITIONALS = True
            # Options
            elif opt in ("--filt-alc-n-points-x-ax"):
                self.FLT_ALC_N_POINTS_X_AX = int(arg)
            elif opt in ("--filt-alc-n-tri-plots"):
                self.FLT_ALC_N_TRI_PLOTS = list(range(int(arg)))
            elif opt in ("--filt-alc-anchor"):
                self.FLT_ALC_ANCHOR = [float(s) for s in arg.split(',')]
            elif opt in ("--filt-alc-range"):
                self.FLT_ALC_RANGE = [float(s) for s in arg.split(',')]

            # Aligned marginals
            elif opt in ("--filtering-marginals"):
                self.FLT_ALIGNED_MARGINALS = True
            # Options
            elif opt in ("--filt-alm-n-points"):
                self.FLT_ALM_N_POINTS = int(arg)
            elif opt in ("--filt-alm-n-tri-plots"):
                self.FLT_ALM_N_TRI_PLOTS = int(arg)

            # Quadrature
            elif opt in ("--filtering-quadrature"):
                self.FLT_QUADRATURE.append( True )
                self.FLT_QUAD_QTYPE.append( None )
                self.FLT_QUAD_QNUM.append( None )
            elif opt in ("--filt-quad-qtype"):
                self.FLT_QUAD_QTYPE[len(self.FLT_QUADRATURE)-1] = int(arg)
            elif opt in ("--filt-quad-qnum"):
                self.FLT_QUAD_QNUM[len(self.FLT_QUADRATURE)-1] = [int(q) for q in arg.split(',')]

    def __init__(self, argv):
        self.TRIM = None
        # Sequential diagnostics
        self.SEQUENTIAL_VAR_DIAG = False
        self.SEQUENTIAL_REG_DIAG = False
        # Filtering aligned conditionals
        self.FLT_ALIGNED_CONDITIONALS = False
        self.FLT_ALC_N_POINTS_X_AX = 30
        self.FLT_ALC_N_TRI_PLOTS = 0
        self.FLT_ALC_ANCHOR = None
        self.FLT_ALC_RANGE = [-5.,5.]
        # Filtering Aligned marginals
        self.FLT_ALIGNED_MARGINALS = False
        self.FLT_ALM_N_POINTS = 1000
        self.FLT_ALM_N_TRI_PLOTS = 0
        # Samples
        self.FLT_QUADRATURE = []
        self.FLT_QUAD_QTYPE = []
        self.FLT_QUAD_QNUM = []

        super(SequentialPostprocess, self).__init__(argv)

    def load(self):
        super(SequentialPostprocess, self).load()
        if self.TRIM is None:
            self.TRIM = self.target_distribution.nsteps
        if self.TRIM == self.target_distribution.nsteps:
            self.filt_tmap_list = self.stg.integrator.filtering_map_list
        elif self.TRIM < self.target_distribution.nsteps:
            integrator = self.stg.integrator.trim(self.TRIM)
            self.target_distribution = self.target_distribution.trim(self.TRIM)
            self.dim = self.target_distribution.dim
            self.base_distribution = DIST.StandardNormalDistribution(self.dim)
            self.tmap = integrator.smoothing_map
            self.approx_base_distribution = DIST.PullBackTransportMapDistribution(
                self.tmap, self.target_distribution)
            self.approx_target_distribution = DIST.PushForwardTransportMapDistribution(
                self.tmap, self.base_distribution)
        else:
            raise ValueError("The value --trim exceed the total number of steps.")
        # Set hdf5 root
        ROOT_NAME = "trim-%d" % self.TRIM
        if ROOT_NAME not in self.h5_file:
            self.h5_file.create_group(ROOT_NAME)
        self.h5_root = self.h5_file[ROOT_NAME]

    def sequential_variance_diagnostic(self, mpi_pool):
        if self.SEQUENTIAL_VAR_DIAG:
            self.filter_tstamp_print("[Start] Sequential variance diagnostic")
            if self.PLOTTING:
                import matplotlib.pyplot as plt
                fig = plt.figure()
                ax = fig.add_subplot(111)
                ax2 = ax.twinx()
                ax.semilogy(self.stg.integrator.var_diag_convergence)
                ax.set_xlabel("Step")
                ax.set_ylabel(r"$\mathbb{V}[\log\rho/T_i^\sharp\pi_i]$")
                n_coeffs = [ tm.n_coeffs for tm in self.stg.integrator.R_list ]
                ax2.semilogy(n_coeffs, 'k')
                ax2.set_ylabel("number of coefficients")
                if self.STORE_FIG_DIR is not None:
                    self.store_figure(
                        fig, self.STORE_FIG_DIR + "/" + self.TITLE + \
                        '-sequential-var-diag' + self.EXTRA_TIT)
                else:
                    plt.show(False)
            self.filter_tstamp_print("[Stop] Sequential variance diagnostic")

    def sequential_regression_diagnostic(self, mpi_pool):
        if self.SEQUENTIAL_REG_DIAG:
            self.filter_tstamp_print("[Start] Sequential regression diagnostic")
            if self.PLOTTING:
                import matplotlib.pyplot as plt
                fig = plt.figure()
                ax = fig.add_subplot(111)
                ax.semilogy(self.stg.integrator.regression_convergence)
                ax.set_xlabel("Iteration")
                ax.set_ylabel(r"$\Vert H_i - H_{i-1}\circ\tilde{H}_i\Vert_2$")
                if self.STORE_FIG_DIR is not None:
                    self.store_figure(
                        fig, self.STORE_FIG_DIR + "/" + self.TITLE + \
                        '-sequential-reg-diag' + self.EXTRA_TIT)
                else:
                    plt.show(False)
            self.filter_tstamp_print("[Stop] Sequential regression diagnostic")

    def filtering_aligned_conditionals(self, mpi_pool):
        if self.FLT_ALIGNED_CONDITIONALS:
            self.filter_tstamp_print("[Start] Filtering conditionals")
            for n, filt_tmap in enumerate(self.filt_tmap_list):
                self.filter_tstamp_print("        Filtering conditionals " + \
                                         "- Step %d" % n)
                DATA_FIELD = 'filtering-conditionals-%d' % n
                data = self.postproc_data.get(DATA_FIELD, None)
                if data is None:
                    base_density = DIST.StandardNormalDistribution(filt_tmap.dim)
                    d = DIST.PushForwardTransportMapDistribution(
                        filt_tmap, base_density)
                    data = DIAG.computeAlignedConditionals(
                        d, dimensions_vec=self.FLT_ALC_N_TRI_PLOTS,
                        numPointsXax=self.FLT_ALC_N_POINTS_X_AX,
                        pointEval=self.FLT_ALC_ANCHOR, range_vec=self.FLT_ALC_RANGE,
                        mpi_pool=mpi_pool)
                    self.postproc_data[DATA_FIELD] = data
                    if self.OUTPUT is not None:
                        self.safe_store(self.postproc_data, self.OUTPUT)
                if self.PLOTTING:
                    fig = DIAG.plotAlignedConditionals(
                        data=data, show_flag=(self.STORE_FIG_DIR is None))
                    if self.STORE_FIG_DIR is not None:
                        self.store_figure(fig, self.STORE_FIG_DIR + '/' + \
                                          self.TITLE + \
                                          '-filtering-conditionals-%d' % n +\
                                          self.EXTRA_TIT)
            self.filter_tstamp_print("[Stop]  Filtering conditionals")

    def filtering_aligned_marginals(self, mpi_pool):
        if self.FLT_ALIGNED_MARGINALS:
            self.filter_tstamp_print("[Start] Filtering marginals")
            F_GRP_NAME = "/filtering"
            if F_GRP_NAME not in self.h5_root:
                self.h5_root.create_group(F_GRP_NAME)
            fgrp = self.h5_root[F_GRP_NAME]
            for n, filt_tmap in enumerate(self.filt_tmap_list):
                self.filter_tstamp_print("        Filtering marginals " + \
                                         "- Step %d - Sample generation" % n)
                dim = filt_tmap.dim
                # Load values if any
                S_GRP_NAME = "step-%d" % n
                if S_GRP_NAME not in fgrp:
                    fgrp.create_group(S_GRP_NAME)
                sgrp = fgrp[S_GRP_NAME]
                Q_GRP_NAME = "quadrature"
                if Q_GRP_NAME not in sgrp:
                    sgrp.create_group(Q_GRP_NAME)
                qgrp = sgrp[Q_GRP_NAME]
                DSET_NAME = '0'
                if DSET_NAME not in qgrp:
                    qgrp.create_dataset(
                        DSET_NAME, (0,dim), maxshape=(None,dim), dtype='d')
                loaded_samp = qgrp[DSET_NAME]
                if self.FLT_ALM_N_POINTS > loaded_samp.shape[0]:
                    nold = loaded_samp.shape[0]
                    new_nsamp = self.FLT_ALM_N_POINTS - nold
                    base_density = DIST.StandardNormalDistribution(dim)
                    d = DIST.PushForwardTransportMapDistribution(
                        filt_tmap, base_density)
                    x = d.rvs(new_nsamp, mpi_pool=mpi_pool)
                    loaded_samp.resize(self.FLT_ALM_N_POINTS, axis=0)
                    loaded_samp[nold:,:] = x
                self.filter_tstamp_print("        Filtering marginals " + \
                                    "- Step %d - Plotting" % n)
                if self.PLOTTING:
                    fig = DIAG.plotAlignedMarginals(
                        loaded_samp[:self.FLT_ALM_N_POINTS,:], self.FLT_ALM_N_TRI_PLOTS,
                        show_flag=(self.STORE_FIG_DIR is None))
                    if self.STORE_FIG_DIR is not None:
                        self.store_figure(
                            fig, self.STORE_FIG_DIR+'/'+self.TITLE + \
                            '-filtering-marginals-%d' % n + \
                            self.EXTRA_TIT)
            self.filter_tstamp_print("[Stop]  Filtering marginals")

    def filtering_quadratures(self, mpi_pool):
        for _, qtype, qnum in zip(
                self.FLT_QUADRATURE, self.FLT_QUAD_QTYPE, self.FLT_QUAD_QNUM):
            self.filter_tstamp_print("[Start] Quadrature " + str(qtype))
            F_GRP_NAME = "/filtering"
            if F_GRP_NAME not in self.h5_root:
                self.h5_root.create_group(F_GRP_NAME)
            fgrp = self.h5_root[F_GRP_NAME]
            for n, filt_tmap in enumerate(self.filt_tmap_list):
                self.filter_tstamp_print("        Quadrature " + str(qtype) + \
                                         "- Step %d - Sample generation" % n)
                dim = filt_tmap.dim
                # Load values if any
                S_GRP_NAME = "step-%d" % n
                if S_GRP_NAME not in fgrp:
                    fgrp.create_group(S_GRP_NAME)
                sgrp = fgrp[S_GRP_NAME]
                Q_GRP_NAME = "quadrature"
                if Q_GRP_NAME not in sgrp:
                    sgrp.create_group(Q_GRP_NAME)
                qgrp = sgrp[Q_GRP_NAME]
                if qtype == 0: # Monte-Carlo
                    DSET_NAME = str(qtype)
                    if DSET_NAME not in qgrp:
                        qgrp.create_dataset(
                            DSET_NAME, (0,dim), maxshape=(None,dim), dtype='d')
                    loaded_samp = qgrp[DSET_NAME]
                    if qnum[0] > loaded_samp.shape[0]:
                        nold = loaded_samp.shape[0]
                        new_nsamp = qnum[0] - nold
                        base_density = DIST.StandardNormalDistribution(dim)
                        d = DIST.PushForwardTransportMapDistribution(
                            filt_tmap, base_density)
                        x = d.rvs(new_nsamp, mpi_pool=mpi_pool)
                        loaded_samp.resize(qnum[0], axis=0)
                        loaded_samp[nold:,:] = x
                elif qtype == 3: # Gauss quadrature
                    QTYPE_NAME = str(qtype)
                    X_NAME = 'x'
                    W_NAME = 'w'
                    if QTYPE_NAME not in qgrp:
                        qgrp.create_group(QTYPE_NAME)
                    qtp_grp = qgrp[QTYPE_NAME]
                    QNUM_NAME = str(qnum)
                    if QNUM_NAME not in qtp_grp:
                        qtp_grp.create_group(QNUM_NAME)
                        qngrp = qtp_grp[QNUM_NAME]
                        base_density = DIST.StandardNormalDistribution(dim)
                        d = DIST.PushForwardTransportMapDistribution(
                            filt_tmap, base_density)
                        (x, w) = d.quadrature(qtype, qnum, mpi_pool=mpi_pool)
                        qngrp.create_dataset(X_NAME, data=x)
                        qngrp.create_dataset(W_NAME, data=w)
            self.filter_tstamp_print("[Stop]  Quadrature")

    def run(self, mpi_pool):
        super(SequentialPostprocess, self).run(mpi_pool)
        self.sequential_variance_diagnostic(mpi_pool)
        self.sequential_regression_diagnostic(mpi_pool)
        self.filtering_aligned_conditionals(mpi_pool)
        self.filtering_aligned_marginals(mpi_pool)
        self.filtering_quadratures(mpi_pool)